"""Helper functions."""

from dataclasses import dataclass
from importlib import import_module
from numbers import Integral
import shutil
from typing import (
    Any,
    Callable,
    cast,
    Iterable,
    List,
    Literal,
    Optional,
    overload,
    Sequence,
    Union,
)

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components
import scipy.spatial

from morfeus.data import (
    atomic_numbers,
    atomic_symbols,
    cov_radii_pyykko,
    radii_alvarez,
    radii_bondi,
    radii_crc,
    radii_rahm,
    radii_truhlar,
)
from morfeus.typing import Array2D, ArrayLike1D, ArrayLike2D


def get_excluded_from_connectivity(
    connectivity_matrix: ArrayLike2D,
    center_atoms: ArrayLike1D,
    connected_atoms: ArrayLike1D,
) -> List[int]:
    """Get atom indices to exclude bassed on connectivity and fragmentation.

    Convenience function that determines atoms to exclude from a calculation of a larger
    structure with multiple fragments. Connected atoms belong to the fragment of
    interest, e.g., a ligand. Center atoms are those of e.g. a central metal atom that.
    By default, the center atoms are added to the excluded ones.

    Args:
        connectivity_matrix: Connectivity matrix
        center_atoms: Atoms of central unit which connects to fragment (1-indexed)
        connected_atoms: Atoms of fragment (1-indexed)

    Returns:
        excluded_atoms: Atom indices to exclude

    Raises:
        ValueError: When connected atoms belong to different fragments or when connected
            atoms belong to same fragment as other neighbors of center atoms (1-indexed)
    """
    connectivity_matrix = np.array(connectivity_matrix)
    center_atoms = np.array(center_atoms).reshape(-1) - 1
    connected_atoms = np.array(connected_atoms).reshape(-1) - 1
    # Determine other neihgbors to the central atoms
    other_neighbors = set(
        connectivity_matrix[center_atoms].reshape(-1).nonzero()[0]
    ).difference(connected_atoms)

    # Calculate fragment labels
    mask = np.ones(len(connectivity_matrix), dtype=bool)
    mask[center_atoms] = False
    graph = csr_matrix(connectivity_matrix)[mask, :][:, mask]
    n_components, labels = connected_components(
        csgraph=graph, directed=False, return_labels=True
    )

    # Take out labels and check for errors
    connected_labels = set([labels[i] for i in connected_atoms])
    if len(connected_labels) > 1:
        raise ValueError("Connected atoms belong to different fragments.")
    neighbor_labels = set([labels[i] for i in other_neighbors])
    if len(neighbor_labels.intersection(connected_labels)) > 0:
        raise ValueError(
            "Connected atoms belong to same fragment as other neighbor of center atoms."
        )
    ref_label = list(connected_labels)[0]
    excluded_atoms = list(np.where(labels != ref_label)[0] + 1)

    return excluded_atoms


def check_distances(
    elements: Union[Iterable[int], Iterable[str]],
    coordinates: ArrayLike2D,
    check_atom: int,
    radii: Optional[ArrayLike1D] = None,
    check_radius: float = 0,
    excluded_atoms: Optional[Sequence[int]] = None,
    epsilon: float = 0,
    radii_type: str = "crc",
) -> List[int]:
    """Check which atoms are within clashing vdW radii distances.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        check_atom: Index of atom to check against (1-indexed)
        radii: vdW radii (Å)
        check_radius: Radius to use for check_atom (Å)
        excluded_atoms: Atom indices to exclude (1-indexed)
        epsilon: Numeric term add to the radii (Å)
        radii_type: Radii type: 'alvarez', 'bondi', 'crc', 'pyykko', 'rahm' or 'truhlar'

    Returns:
        within_list: Atom indices within vdW distance of check atom.
    """
    # Convert elements to atomic numbers if the are symbols
    elements = convert_elements(elements, output="numbers")

    # Get radii if they are not supplied
    if radii is None:
        radii = get_radii(elements, radii_type=radii_type)
    radii = np.array(radii)

    if excluded_atoms is None:
        excluded_atoms = []
    else:
        excluded_atoms = list(excluded_atoms)

    coordinates = np.array(coordinates)
    atom_coordinates = np.array(coordinates)
    check_coordinates = np.array(coordinates[check_atom - 1]).reshape(-1, 3)

    # Calculate distances between check atom and all atoms
    distances = (
        scipy.spatial.distance.cdist(atom_coordinates, check_coordinates)
        - radii.reshape(-1, 1)
        - check_radius
        - epsilon
    )
    distances = distances.reshape(-1)

    # Determine atoms which are within a vdW distance from the check atom
    within_distance = list(np.argwhere(distances < 0).reshape(-1))

    # Remove check atom and atoms in the exclude list
    within_distance.remove(check_atom - 1)
    within_distance = [i + 1 for i in within_distance if i + 1 not in excluded_atoms]

    return within_distance


def requires_executable(executables: Sequence[str]) -> Callable[..., Callable]:
    """Decorator factory to control optional executables.

    Args:
        executables: Names of executables

    Returns:
        decorator: Either 'noop_decorator' that returns the original function or
            'error_decorator' that raises an OSError and lists absent executables.
    """

    def noop_decorator(function: Callable[..., Any]) -> Callable[..., Any]:
        """Returns function unchanged."""
        return function

    def error_decorator(function: Callable[..., Any]) -> Callable[..., Any]:
        """Raises error."""

        def error(*args, **kwargs) -> OSError:
            error_msg = "Required executables not found in path:"
            for exe_error in exe_errors:
                error_msg += f" {exe_error}"
            raise OSError(error_msg)

        return error

    # Try to find excetubles in path
    exe_errors = []
    for executable in executables:
        if not shutil.which(executable):
            exe_errors.append(executable)

    return error_decorator if len(exe_errors) > 0 else noop_decorator


@dataclass
class Import:
    """Class for handling optional dependency imports."""

    module: str
    item: Optional[str] = None
    alias: Optional[str] = None


def requires_dependency(  # noqa: C901
    imports: Sequence[Import], _globals: dict
) -> Callable[..., Callable]:
    """Decorator factory to control optional dependencies.

    Args:
        imports: Imports
        _globals: Global symbol table from calling module.

    Returns:
        decorator: Either 'noop_decorator' that returns the original function or
            'error_decorator' that raises an ImportError and lists absent dependencies.
    """

    def noop_decorator(function: Callable[..., Any]) -> Callable[..., Any]:
        """Returns function unchanged."""
        return function

    def error_decorator(function: Callable[..., Any]) -> Callable[..., Any]:
        """Raises error."""

        def error(*args, **kwargs) -> ImportError:
            error_msg = "Install extra requirements to use this function:"
            for e in import_errors:
                error_msg += f" {e.name}"
            raise ImportError(error_msg)

        return error

    import_errors = []
    for imp in imports:
        # Import module
        try:
            module = import_module(imp.module)

            # Try to import item as attribute
            if imp.item is not None:
                try:
                    item = getattr(module, imp.item)
                except AttributeError:
                    item = import_module(f"{imp.module}.{imp.item}")
                name = imp.item
            else:
                item = module
                name = imp.module

            # Convert item name to alias
            if imp.alias is not None:
                name = imp.alias

            _globals[name] = item
        except ImportError as import_error:
            import_errors.append(import_error)

    return error_decorator if len(import_errors) > 0 else noop_decorator


@overload
def convert_elements(
    elements: Union[Iterable[int], Iterable[str]], output: Literal["numbers"]
) -> List[int]:
    ...


@overload
def convert_elements(
    elements: Union[Iterable[int], Iterable[str]], output: Literal["symbols"]
) -> List[str]:
    ...


def convert_elements(
    elements: Union[Iterable[int], Iterable[str]], output: str = "numbers"
) -> Union[List[int], List[str]]:
    """Converts elements to atomic symbols or numbers.

    Args:
        elements: Elements as atomic symbols or numbers
        output: Output format: 'numbers' (default) or 'symbols'.

    Returns:
        elements: Converted elements

    Raises:
        TypeError: When input type not supported
        ValueError: When output not supported
    """
    if output not in ["numbers", "symbols"]:
        raise ValueError(f"ouput={output} not supported. Use 'numbers' or 'symbols'")

    if all(isinstance(element, str) for element in elements):
        elements = cast(List[str], elements)
        if output == "numbers":
            elements = [atomic_numbers[element.capitalize()] for element in elements]
        return elements
    elif all(isinstance(element, Integral) for element in elements):
        elements = cast(List[int], elements)
        if output == "symbols":
            elements = [atomic_symbols[element] for element in elements]
        return elements
    else:
        raise TypeError("elements must be all integers or all strings.")


def get_radii(
    elements: Union[Iterable[int], Iterable[str]],
    radii_type: str = "crc",
    scale: float = 1,
) -> List[float]:
    """Gets radii from element identifiers.

    Args:
        elements: Elements as atomic symbols or numbers
        radii_type: Radii type: 'alvarez', 'bondi', 'crc', 'pyykko' 'rahm' or 'truhlar'
        scale: Scaling factor

    Returns:
        radii: Radii (Å)
    """
    elements = convert_elements(elements, output="numbers")

    # Set up dictionary of radii types
    radii_choice = {
        "alvarez": radii_alvarez,
        "bondi": radii_bondi,
        "crc": radii_crc,
        "rahm": radii_rahm,
        "pyykko": cov_radii_pyykko,
        "truhlar": radii_truhlar,
    }

    # Get the radii. Replace with 2.0 if it the radius doesn't exist.
    radii = [radii_choice[radii_type].get(element, 2.0) * scale for element in elements]

    return radii


def get_connectivity_matrix(
    coordinates: ArrayLike2D,
    elements: Optional[Union[Iterable[int], Iterable[str]]] = None,
    radii: Optional[ArrayLike1D] = None,
    radii_type: str = "pyykko",
    scale_factor: float = 1.2,
) -> Array2D:
    """Get connectivity matrix from covalent radii.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        radii: Radii (Å)
        radii_type: Radii type: 'pyykko'
        scale_factor: Factor for scaling covalent radii

    Returns:
        connectivity_matrix: Connectivity matrix

    Raises:
        RuntimeError: When neither elements nor radii given
    """
    coordinates = np.array(coordinates)
    n_atoms = len(coordinates)
    if radii is None:
        if elements is None:
            raise RuntimeError("Either elements or radii needed.")
        elements = convert_elements(elements, output="numbers")
        radii = get_radii(elements, radii_type=radii_type)
    radii = np.array(radii)
    distance_matrix = scipy.spatial.distance_matrix(coordinates, coordinates)
    radii_matrix = np.add.outer(radii, radii) * scale_factor
    connectivity_matrix: np.ndarray = (distance_matrix < radii_matrix) - np.identity(
        n_atoms
    ).astype(int)

    return connectivity_matrix



"""Dispersion code."""

import functools
from os import PathLike
import typing
from typing import Any, cast, Dict, Iterable, List, Optional, Sequence, Union

import numpy as np
import scipy.spatial

from morfeus.calculators import D3Calculator, D3Grimme, D4Grimme
from morfeus.data import ANGSTROM_TO_BOHR, atomic_symbols, HARTREE_TO_KCAL, jmol_colors
from morfeus.geometry import Atom
from morfeus.io import CubeParser, D3Parser, D4Parser, read_geometry, VertexParser
from morfeus.sasa import SASA
from morfeus.typing import Array1D, Array2D, ArrayLike1D, ArrayLike2D
from morfeus.utils import convert_elements, get_radii, Import, requires_dependency

if typing.TYPE_CHECKING:
    from matplotlib.colors import hex2color
    import pymeshfix
    import pyvista as pv
    from pyvistaqt import BackgroundPlotter
    import vtk


class Dispersion:
    """Calculates and stores the results for the 🍺P_int dispersion descriptor.

    The descriptor is defined in 10.1002/anie.201905439. Morfeus can compute it based on
    a surface either from vdW radii, surface vertices or the electron density.
    Dispersion can be obtained with the D3 or D4 model.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        radii: VdW radii (Å)
        radii_type: Choice of vdW radii: 'alvarez', 'bondi', 'crc', 'rahm' and 'truhlar'
        point_surface: Use point surface from vdW radii
        compute_coefficients: Whether to compute D3 coefficients with internal code
        density: Area per point (Å²) on the vdW surface
        excluded_atoms: Atoms to exclude (1-indexed). Used for substituent P_ints
        included_atoms: Atoms to include. Used for functional group P_ints

    Attributes:
        area: Area of surface (Å²)
        atom_areas: Atom indices as keys and atom areas as values (Å²)
        atom_p_int: Atom indices as keys and P_int as values (kcal¹ᐟ² mol⁻¹⸍²))
        atom_p_max: Atom indices as keys and P_max as values (kcal¹ᐟ² mol⁻¹ᐟ²)
        atom_p_min: Atom indices as keys and P_min as values( kcal¹ᐟ² mol⁻¹ᐟ²)
        p_int: P_int value for molecule (kcal¹ᐟ² mol⁻¹ᐟ²)
        p_max: Highest P value (kcal¹ᐟ² mol⁻¹ᐟ²)
        p_min: Lowest P value (kcal¹ᐟ² mol⁻¹ᐟ²)
        p_values: All P values (kcal¹ᐟ² mol⁻¹ᐟ²)
        volume: Volume of surface (Å³)

    Raises:
        Exception: When both exluded_atoms and included_atom are given
    """

    area: float
    atom_areas: Dict[int, float]
    atom_p_int: Dict[int, float]
    atom_p_max: Dict[int, float]
    atom_p_min: Dict[int, float]
    p_int: float
    p_max: float
    p_min: float
    p_values: Array1D
    volume: float
    _atoms: List[Atom]
    _c_n_coefficients: Dict[int, Array1D]
    _density: float
    _excluded_atoms: List[int]
    _point_areas: Array1D
    _point_map: Array1D
    _points: Array2D
    _radii: Array1D
    _surface: "pv.PolyData"

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        radii: Optional[ArrayLike1D] = None,
        radii_type: str = "rahm",
        point_surface: bool = True,
        compute_coefficients: bool = True,
        density: float = 0.1,
        excluded_atoms: Optional[Sequence[int]] = None,
        included_atoms: Optional[Sequence[int]] = None,
    ) -> None:
        # Check that only excluded or included atoms are given
        if excluded_atoms is not None and included_atoms is not None:
            raise Exception("Give either excluded or included atoms but not both.")

        # Converting elements to atomic numbers if the are symbols
        elements = convert_elements(elements, output="numbers")
        coordinates = np.array(coordinates)

        # Set excluded atoms
        all_atoms = set(range(1, len(elements) + 1))
        if included_atoms is not None:
            included_atoms_ = set(included_atoms)
            excluded_atoms = list(all_atoms - included_atoms_)
        elif excluded_atoms is None:
            excluded_atoms = []
        else:
            excluded_atoms = list(excluded_atoms)

        self._excluded_atoms = excluded_atoms

        # Set up
        self._surface = None
        self._density = density

        # Getting radii if they are not supplied
        if radii is None:
            radii = get_radii(elements, radii_type=radii_type)
        radii = np.array(radii)

        self._radii = radii

        # Get vdW surface if requested
        if point_surface:
            self._surface_from_sasa(elements, coordinates)
        else:
            # Get list of atoms as Atom objects
            atoms: List[Atom] = []
            for i, (element, coord, radius) in enumerate(
                zip(elements, coordinates, radii), start=1
            ):
                atom = Atom(element, coord, radius, i)
                atoms.append(atom)
            self._atoms = atoms

        # Calculate coefficients
        if compute_coefficients:
            self.compute_coefficients(model="id3")

        # Calculatte P_int values
        if point_surface and compute_coefficients:
            self.compute_p_int()

    def _surface_from_sasa(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
    ) -> None:
        """Get surface from SASA."""
        sasa = SASA(
            elements,
            coordinates,
            radii=self._radii,
            density=self._density,
            probe_radius=0,
        )
        self._atoms = sasa._atoms
        self.area = sum(
            [
                atom.area
                for atom in self._atoms
                if atom.index not in self._excluded_atoms
            ]
        )
        self.atom_areas = sasa.atom_areas
        self.volume = sum(
            [
                atom.volume
                for atom in self._atoms
                if atom.index not in self._excluded_atoms
            ]
        )

        # Get point areas and map from point to atom
        point_areas: List[np.ndarray] = []
        point_map = []
        for atom in self._atoms:
            n_points = len(atom.accessible_points)
            if n_points > 0:
                point_area = atom.area / n_points
            else:
                point_area = 0.0
            atom.point_areas = np.repeat(point_area, n_points)
            point_areas.extend(atom.point_areas)
            point_map.extend([atom.index] * n_points)
        self._point_areas = np.array(point_areas)
        self._point_map = np.array(point_map)

    @requires_dependency([Import(module="pyvista", alias="pv")], globals())
    def surface_from_cube(
        self,
        file: Union[str, PathLike],
        isodensity: float = 0.001,
        method: str = "flying_edges",
    ) -> "Dispersion":
        """Adds an isodensity surface from a Gaussian cube file.

        Args:
            file: Gaussian cube file
            isodensity: Isodensity value (electrons/bohr³)
            method: Method for contouring: 'contour' or 'flying_edges

        Returns:
            self: Self
        """
        # Parse the cubefile
        parser = CubeParser(file)

        # Generate grid and fill with values
        grid = pv.UniformGrid()
        grid.dimensions = np.array(parser.X.shape)
        grid.origin = (parser.min_x, parser.min_y, parser.min_z)
        grid.spacing = (parser.step_x, parser.step_y, parser.step_z)
        grid.point_arrays["values"] = parser.S.flatten(order="F")
        self.grid = grid

        # Contour and process the surface
        surface = self._contour_surface(grid, method=method, isodensity=isodensity)
        self._surface = surface
        self._process_surface()

        return self

    @requires_dependency(
        [Import("pymeshfix"), Import(module="pyvista", alias="pv")], globals()
    )
    def surface_from_multiwfn(
        self, file: Union[str, PathLike], fix_mesh: bool = True
    ) -> "Dispersion":
        """Adds surface from Multiwfn vertex file with connectivity information.

        Args:
            file: Vertex.pdb file
            fix_mesh: Whether to fix holes in the mesh with pymeshfix (recommended)

        Returns:
            self: Self
        """
        # Read the vertices and faces from the Multiwfn output file
        parser = VertexParser(file)
        vertices = np.array(parser.vertices)
        faces = np.array(parser.faces)
        faces = np.insert(faces, 0, values=3, axis=1)

        # Construct surface and fix it with pymeshfix
        surface = pv.PolyData(vertices, faces, show_edges=True)
        if fix_mesh:
            meshfix = pymeshfix.MeshFix(surface)
            meshfix.repair()
            surface = meshfix.mesh

        # Process surface
        self._surface = surface
        self._process_surface()

        return self

    def _process_surface(self) -> None:
        """Extracts face center points and assigns these to atoms based on proximity."""
        # Get the area and volume
        self.area = self._surface.area
        self.volume = self._surface.volume

        # Assign face centers to atoms according to Voronoi partitioning
        coordinates = np.array([atom.coordinates for atom in self._atoms])
        points = np.array(self._surface.cell_centers().points)
        kd_tree = scipy.spatial.cKDTree(coordinates)
        _, point_regions = kd_tree.query(points, k=1)
        point_regions = point_regions + 1

        # Compute faces areas
        area_data = self._surface.compute_cell_sizes()
        areas = np.array(area_data.cell_arrays["Area"])

        # Assign face centers and areas to atoms
        atom_areas = {}
        for atom in self._atoms:
            atom.accessible_points = points[point_regions == atom.index]
            point_areas = areas[point_regions == atom.index]
            atom.area = np.sum(point_areas)
            atom.point_areas = point_areas
            atom_areas[atom.index] = atom.area

        # Set up attributes
        self.atom_areas = atom_areas
        self._point_areas = areas
        self._point_map = point_regions

    @requires_dependency(
        [Import(module="pyvista", alias="pv"), Import("vtk")], globals()
    )
    @staticmethod
    def _contour_surface(
        grid: "pv.Grid", method: str = "flying_edges", isodensity: float = 0.001
    ) -> "pv.PolyData":
        """Counter surface from grid.

        Args:
            grid: Electron density as PyVista Grid object
            isodensity: Isodensity value (electrons/bohr³)
            method: Method for contouring: 'contour' or 'flying_edges

        Returns:
            surface: Surface as Pyvista PolyData object
        """
        # Select method for contouring
        if method == "flying_edges":
            contour_filter = vtk.vtkFlyingEdges3D()
        elif method == "contour":
            contour_filter = vtk.vtkContourFilter()

        # Run the contour filter
        isodensity = isodensity
        contour_filter.SetInputData(grid)
        contour_filter.SetValue(0, isodensity)
        contour_filter.Update()
        surface = contour_filter.GetOutput()
        surface = pv.wrap(surface)

        return surface

    def compute_p_int(self, points: Optional[ArrayLike2D] = None) -> "Dispersion":
        """Compute P_int values for surface or points.

        Args:
            points: Points to compute P values for

        Returns:
            self: Self
        """
        # Set up atoms and coefficients that are part of the calculation
        atom_indices = np.array(
            [
                atom.index - 1
                for atom in self._atoms
                if atom.index not in self._excluded_atoms
            ]
        )
        coordinates = np.array([atom.coordinates for atom in self._atoms])
        coordinates = coordinates[atom_indices]
        c_n_coefficients = dict(self._c_n_coefficients)
        for key, value in c_n_coefficients.items():
            c_n_coefficients[key] = np.array(value)[atom_indices] * HARTREE_TO_KCAL

        # Take surface points if none are given
        if points is None:
            points = np.vstack(
                [
                    atom.accessible_points
                    for atom in self._atoms
                    if atom.index not in self._excluded_atoms
                    and atom.accessible_points.size > 0
                ]
            )
            atomic = True
        else:
            points = np.array(points)

        # Calculate p_int for each point
        dist = scipy.spatial.distance.cdist(points, coordinates) * ANGSTROM_TO_BOHR
        p = np.sum(
            [
                np.sum(np.sqrt(coefficients / (dist ** order)), axis=1)
                for order, coefficients in c_n_coefficients.items()
            ],
            axis=0,
        )
        p = cast(np.ndarray, p)

        self.p_values = p

        # Take out atomic p_ints if no points are given
        if atomic:
            atom_p_max = {}
            atom_p_min = {}
            atom_p_int = {}
            i_start = 0
            for atom in self._atoms:
                if atom.index not in self._excluded_atoms:
                    n_points = len(atom.accessible_points)
                    if n_points > 0:
                        i_stop = i_start + n_points
                        atom_ps = p[i_start:i_stop]
                        atom.p_values = atom_ps
                        atom_p_max[atom.index] = np.max(atom_ps)
                        atom_p_min[atom.index] = np.min(atom_ps)
                        atom_p_int[atom.index] = np.sum(
                            atom_ps * atom.point_areas / atom.area
                        )
                        i_start = i_stop
                    else:
                        atom_p_max[atom.index] = 0
                        atom_p_min[atom.index] = 0
                        atom_p_int[atom.index] = 0
                        atom.p_values = np.array([])
            self.atom_p_max = atom_p_max
            self.atom_p_min = atom_p_min
            self.atom_p_int = atom_p_int

        point_areas = self._point_areas[np.isin(self._point_map, atom_indices + 1)]
        self.p_int = np.sum(p * point_areas / self.area)

        # Calculate p_min and p_max with slight modification to Robert's
        # definitions
        self.p_min = np.min(p)
        self.p_max = np.max(p)

        # Map p_values onto surface
        if self._surface:
            mapped_p = np.zeros(len(p))
            for atom in self._atoms:
                if atom.index not in self._excluded_atoms:
                    mapped_p[self._point_map == atom.index] = atom.p_values
            self._surface.cell_arrays["values"] = mapped_p
            self._surface = self._surface.cell_data_to_point_data()

        # Store points for later use
        self._points = points

        return self

    def compute_coefficients(
        self, model: str = "id3", order: int = 8, charge: int = 0
    ) -> "Dispersion":
        """Compute dispersion coefficients.

        Can either use internal D3 model or D4 or D3-like model available through
        Grimme's dftd4 program.

        Args:
            model: Calculation model: 'id3'. 'gd3' or 'gd4'
            order: Order of the Cᴬᴬ coefficients
            charge: Molecular charge for D4 model

        Returns:
            self: Self

        Raises:
            ValueError: When model not supported
        """
        # Set up atoms and coordinates
        elements = [atom.element for atom in self._atoms]
        coordinates = np.array([atom.coordinates for atom in self._atoms])

        calculators = {
            "id3": D3Calculator,
            "gd3": D3Grimme,
            "gd4": D4Grimme,
        }
        calc: Union[D3Calculator, D3Grimme, D4Grimme]
        # Calculate  D3 values with internal model
        if model in ["id3", "gd3"]:
            calc = calculators[model](elements, coordinates, order=order)
        elif model in ["gd4"]:
            calc = calculators[model](elements, coordinates, order=order, charge=charge)
        else:
            raise ValueError(f"model={model} not supported.")
        self._c_n_coefficients = calc.c_n_coefficients

        return self

    def load_coefficients(self, file: Union[str, PathLike], model: str) -> "Dispersion":
        """Load the C₆ and C₈ coefficients.

        Output can be read from the dftd3 and dftd4 programs by giving a file in
        combination with the corresponding model.

        Args:
            file: Output file from the dftd3 or dftd4 programs
            model: Calculation model: 'd3' or 'd4'

        Returns:
            self: Self

        Raises:
            ValueError: When model not supported
        """
        parser: Union[D3Parser, D4Parser]
        if model == "d3":
            parser = D3Parser(file)
        elif model == "d4":
            parser = D4Parser(file)
        else:
            raise ValueError(f"model={model} not supported.")
        self._c_n_coefficients = {}
        self._c_n_coefficients[6] = parser.c6_coefficients
        self._c_n_coefficients[8] = parser.c8_coefficients

        return self

    def print_report(self, verbose: bool = False) -> None:
        """Print report of results.

        Args:
            verbose: Whether to print atom P_ints
        """
        print(f"Surface area (Å²): {self.area:.1f}")
        print(f"Surface volume (Å³): {self.volume:.1f}")
        print(f"P_int (kcal¹ᐟ² mol⁻¹ᐟ²): {self.p_int:.1f}")
        if verbose:
            print(
                f"{'Symbol':<10s}{'Index':<10s}{'P_int (kcal^(1/2) mol^(-1/2))':<30s}"
            )
            for atom, (i, p_int) in zip(self._atoms, self.atom_p_int.items()):
                symbol = atomic_symbols[atom.element]
                print(f"{symbol:<10s}{i:<10d}{p_int:<10.1f}")

    def save_vtk(self, filename: str) -> "Dispersion":
        """Save surface as .vtk file.

        Args:
            filename: Name of file. Use .vtk suffix.

        Returns:
            self: Self
        """
        self._surface.save(filename)

        return self

    @requires_dependency(
        [
            Import(module="matplotlib.colors", item="hex2color"),
            Import(module="pyvista", alias="pv"),
            Import(module="pyvistaqt", item="BackgroundPlotter"),
        ],
        globals(),
    )
    def draw_3D(
        self,
        opacity: float = 1,
        display_p_int: bool = True,
        molecule_opacity: float = 1,
        atom_scale: float = 1,
    ) -> None:
        """Draw surface with mapped P_int values.

        Args:
            opacity: Surface opacity
            display_p_int: Whether to display P_int mapped onto the surface
            molecule_opacity: Molecule opacity
            atom_scale: Scale factor for atom size
        """
        # Set up plotter
        p = BackgroundPlotter()

        # Draw molecule
        for atom in self._atoms:
            color = hex2color(jmol_colors[atom.element])
            radius = atom.radius * atom_scale
            sphere = pv.Sphere(center=list(atom.coordinates), radius=radius)
            p.add_mesh(
                sphere, color=color, opacity=molecule_opacity, name=str(atom.index)
            )

        cmap: Optional[str]
        # Set up plotting of mapped surface
        if display_p_int is True:
            color = None
            cmap = "coolwarm"
        else:
            color = "tan"
            cmap = None

        # Draw surface
        if self._surface:
            p.add_mesh(self._surface, opacity=opacity, color=color, cmap=cmap)
        else:
            point_cloud = pv.PolyData(self._points)
            point_cloud["values"] = self.p_values
            p.add_mesh(
                point_cloud,
                opacity=opacity,
                color=color,
                cmap=cmap,
                render_points_as_spheres=True,
            )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self._atoms)!r} atoms)"


def cli(file: str) -> Any:
    """CLI for dispersion descriptor.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    elements, coordinates = read_geometry(file)
    return functools.partial(Dispersion, elements, coordinates)



"""Sterimol code."""

import functools
import math
import typing
from typing import Any, cast, Iterable, List, Optional, Sequence, Set, Union

import numpy as np
import scipy.spatial

from morfeus.data import jmol_colors
from morfeus.geometry import Atom, kabsch_rotation_matrix, sphere_line_intersection
from morfeus.io import read_geometry
from morfeus.plotting import get_drawing_arrow
from morfeus.sasa import SASA
from morfeus.typing import Array1D, ArrayLike1D, ArrayLike2D
from morfeus.utils import convert_elements, get_radii, Import, requires_dependency

if typing.TYPE_CHECKING:
    from matplotlib.colors import hex2color
    import pyvista as pv
    from pyvistaqt import BackgroundPlotter


class Sterimol:
    """Performs and stores results of Sterimol calculation.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        dummy_index: Index of dummy atom (1-indexed)
        attached_index: Index of attached atom of substituent (1-indexed). For a list of
            indices, a dummy atom is created at their geometric center
        radii: List of radii (Å)
        radii_type: vdW radii type: 'alvarez', 'bondi', 'crc' or 'truhlar'
        n_rot_vectors: Number of rotational vectors for determining B₁ and B₅
        excluded_atoms: Atom indices to exclude from the calculation
        calculate: Whether to calculate the Sterimol parameters directly

    Attributes:
        B_1_value: Sterimol B₁ value (Å)
        B_1: Sterimol B₁ vector (Å)
        B_5_value: Sterimol B_5 value (Å)
        B_5: Sterimol B₅ vector (Å)
        bond_length: Bond length between atom 1 and atom 2 (Å)
        L_value_uncorrected: Sterimol L value minus 0.40 (Å)
        L_value: Sterimol L value (Å)
        L: Sterimol L vector (Å)
    """

    B_1_value: float
    B_1: Array1D
    B_5_value: float
    B_5: Array1D
    bond_length: float
    L_value_uncorrected: float
    L_value: float
    L: Array1D
    _atoms: List[Atom]
    _attached_atom: Atom
    _dummy_atom: Atom
    _excluded_atoms: Set[int]
    _n_rot_vectors: int
    _origin: Array1D
    _plotter: "BackgroundPlotter"
    _points: Array1D
    _sphere_radius: float

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        dummy_index: int,
        attached_index: Union[int, Iterable[int]],
        radii: Optional[ArrayLike1D] = None,
        radii_type: str = "crc",
        n_rot_vectors: int = 3600,
        excluded_atoms: Optional[Sequence[int]] = None,
        calculate: bool = True,
    ) -> None:
        # Convert elements to atomic numbers if the are symbols
        elements = convert_elements(elements, output="numbers")
        coordinates = np.array(coordinates)

        if excluded_atoms is None:
            excluded_atoms = []
        excluded_atoms = np.array(excluded_atoms)

        # Get radii if they are not supplied
        if radii is None:
            radii = get_radii(elements, radii_type=radii_type)
        radii = np.array(radii)

        # Add dummy atom if multiple attached indices are given
        if isinstance(attached_index, Iterable):
            attached_dummy_coordinates = np.mean(
                [coordinates[i - 1] for i in attached_index], axis=0
            )
            attached_dummy_coordinates = cast(np.ndarray, attached_dummy_coordinates)
            coordinates = np.vstack([coordinates, attached_dummy_coordinates])
            elements.append(0)
            radii = np.concatenate([radii, [0.0]])
            attached_index = len(elements)

        # Set up coordinate array
        all_coordinates = coordinates
        all_radii = radii

        # Translate coordinates so origin is at atom 2
        origin = all_coordinates[attached_index - 1]
        all_coordinates -= origin

        # Get vector pointing from atom 2 to atom 1
        vector_2_to_1 = (
            all_coordinates[attached_index - 1] - all_coordinates[dummy_index - 1]
        )
        bond_length = np.linalg.norm(vector_2_to_1)
        vector_2_to_1 = vector_2_to_1 / np.linalg.norm(vector_2_to_1)

        # Get rotation quaternion that overlays vector with x-axis
        x_axis = np.array([[1.0, 0.0, 0.0]])
        R = kabsch_rotation_matrix(vector_2_to_1.reshape(1, -1), x_axis, center=False)
        all_coordinates = (R @ all_coordinates.T).T
        self._rotation_matrix = R

        # Get list of atoms as Atom objects
        atoms = []
        for i, (element, radius, coord) in enumerate(
            zip(elements, all_radii, all_coordinates), start=1
        ):
            atom = Atom(element, coord, radius, i)
            atoms.append(atom)
            if i == dummy_index:
                dummy_atom = atom
            if i == attached_index:
                attached_atom = atom

        # Set up attributes
        self._atoms = atoms
        self._excluded_atoms = set(excluded_atoms)
        self._origin = origin

        self._dummy_atom = dummy_atom
        self._attached_atom = attached_atom

        self.bond_length = bond_length

        self._n_rot_vectors = n_rot_vectors

        if calculate:
            self.calculate()

    def set_points(
        self, points: Sequence[Sequence[float]], shift: bool = True
    ) -> "Sterimol":
        """Set points for calculation of Sterimol.

        Args:
            points: Points (Å)
            shift: Whether to shift the points to have origin at dummy atom.

        Returns:
            self: Self
        """
        points = np.array(points)
        if shift is True:
            points -= self._origin

        return self

    def bury(
        self,
        sphere_radius: float = 5.5,
        method: str = "delete",
        radii_scale: float = 0.5,
        density: float = 0.01,
    ) -> "Sterimol":
        """Do a Buried Sterimol calculation.

        There are three available schemes based on deletion, truncation or slicing.

        Args:
            sphere_radius: Radius of sphere (Å)
            method: Method for burying: 'delete', 'slice' or 'truncate'
            radii_scale: Scale radii for metohd='delete' calculation
            density: Area per point on surface (Å²)

        Returns:
            self: Self

        Raises:
            ValueError: When method is not specified correctly
            Exception: When using method='truncate' and sphere is too small
        """
        if method == "delete":
            # Remove all atoms outside sphere (taking vdW radii into account)
            coordinates = np.vstack([atom.coordinates for atom in self._atoms])
            radii = np.array([atom.radius for atom in self._atoms])
            distances = scipy.spatial.distance.cdist(
                self._dummy_atom.coordinates.reshape(1, -1), coordinates
            ).reshape(-1)
            distances = distances - radii * radii_scale
            excluded_atoms = set(np.array(self._atoms)[distances >= sphere_radius])
            self._excluded_atoms.update([atom.index for atom in excluded_atoms])

            # Calculate Sterimol parameters
            self.calculate()
        elif method == "truncate":
            # Calculate Sterimol parameters
            self.calculate()

            # Calculate intersection between vectors and sphere.
            atom_1_coordinates = self._dummy_atom.coordinates
            atom_2_coordinates = self._attached_atom.coordinates

            new_vectors = []
            for vector, ref_coordinates in [
                (self.L, atom_1_coordinates),
                (self.B_1, atom_2_coordinates),
                (self.B_5, atom_2_coordinates),
            ]:
                # Get intersection point
                intersection_points = sphere_line_intersection(
                    vector, atom_1_coordinates, sphere_radius
                )
                if len(intersection_points) < 1:
                    raise Exception("Sphere so small that vectors don't intersect")

                # Get vector pointing in the right direction
                trial_vectors = [
                    point - ref_coordinates for point in intersection_points
                ]
                norm_vector = vector / np.linalg.norm(vector)
                dot_products = [
                    np.dot(norm_vector, trial_vector / np.linalg.norm(trial_vector))
                    for trial_vector in trial_vectors
                ]
                new_vector = trial_vectors[int(np.argmax(dot_products))]
                new_vectors.append(new_vector)

            # Replace vectors if new ones are shorter than old ones
            if np.linalg.norm(self.L) > np.linalg.norm(new_vectors[0]):
                self.L = new_vectors[0]
                L_value = np.linalg.norm(self.L)
                self.L_value = L_value + 0.40
                self.L_value_uncorrected = L_value
            if np.linalg.norm(self.B_1) > np.linalg.norm(new_vectors[1]):
                self.B_1 = new_vectors[1]
                self.B_1_value = np.linalg.norm(self.B_1)
            if np.linalg.norm(self.B_5) > np.linalg.norm(new_vectors[2]):
                self.B_5 = new_vectors[2]
                self.B_5_value = np.linalg.norm(self.B_5)

        elif method == "slice":
            if not hasattr(self, "_points"):
                self.surface_from_radii(density=density)
            # Remove points outside of sphere
            distances = scipy.spatial.distance.cdist(
                self._dummy_atom.coordinates.reshape(1, -1), self._points
            ).reshape(-1)
            self._points = self._points[distances <= sphere_radius]

            # Calculate Sterimol parameters
            self.calculate()
        else:
            raise ValueError(f"Method: {method} is not supported.")

        # Set attributes
        self._sphere_radius = sphere_radius

        return self

    def surface_from_radii(self, density: float = 0.01) -> "Sterimol":
        """Create surface points from vdW surface.

        Args:
            density: Area per point on surface (Å²)

        Returns:
            self: Self
        """
        # Calculate vdW surface for all active atoms
        elements = []
        coordinates = []
        radii = []
        for atom in self._atoms:
            if atom.index not in self._excluded_atoms and atom is not self._dummy_atom:
                elements.append(atom.element)
                coordinates.append(atom.coordinates)
                radii.append(atom.radius)
        elements = np.array(elements)
        coordinates = np.vstack(coordinates)
        radii = radii
        sasa = SASA(elements, coordinates, radii=radii, density=density, probe_radius=0)

        # Take out points of vdW surface
        points = np.vstack(
            [
                atom.accessible_points
                for atom in sasa._atoms
                if atom.index not in self._excluded_atoms
                and atom.accessible_points.size > 0
            ]
        )
        self._points = points

        return self

    def calculate(self) -> "Sterimol":
        """Calculate Sterimol parameters."""
        # Use coordinates and radii if points are not given
        if not hasattr(self, "_points"):
            coordinates = []
            radii = []
            for atom in self._atoms:
                if (
                    atom is not self._dummy_atom
                    and atom.index not in self._excluded_atoms
                ):
                    coordinates.append(atom.coordinates)
                    radii.append(atom.radius)
            coordinates = np.vstack(coordinates)
            radii = np.vstack(radii).reshape(-1)

        # Project coordinates onto vector between atoms 1 and 2
        vector = self._attached_atom.coordinates - self._dummy_atom.coordinates
        bond_length = np.linalg.norm(vector)
        unit_vector = vector / np.linalg.norm(vector)

        if not hasattr(self, "_points"):
            c_values = np.dot(unit_vector.reshape(1, -1), coordinates.T)
            projected = c_values + radii
        else:
            projected = np.dot(unit_vector.reshape(1, -1), self._points.T)

        # Get L as largest projection along the vector
        L_value = np.max(projected) + bond_length
        L = unit_vector * L_value
        L = L.reshape(-1)

        # Get rotation vectors in yz plane
        r = 1
        theta = np.linspace(0, 2 * math.pi, self._n_rot_vectors)
        x = np.zeros(len(theta))
        y = r * np.cos(theta)
        z = r * np.sin(theta)
        rot_vectors = np.column_stack((x, y, z))

        # Project coordinates onto rotation vectors
        if not hasattr(self, "_points"):
            c_values = np.dot(rot_vectors, coordinates.T)
            projected = c_values + radii
        else:
            projected = np.dot(rot_vectors, self._points.T)
        max_c_values = np.max(projected, axis=1)

        # Determine B1 and B5 from the smallest and largest scalar projections
        B_1_value = np.min(max_c_values)
        B_1 = rot_vectors[np.argmin(max_c_values)] * B_1_value

        B_5_value = np.max(max_c_values)
        B_5 = rot_vectors[np.argmax(max_c_values)] * B_5_value

        # Set up attributes
        self.L = L
        self.L_value = L_value + 0.40
        self.L_value_uncorrected = L_value

        self.B_1 = B_1
        self.B_1_value = B_1_value

        self.B_5 = B_5
        self.B_5_value = B_5_value

        return self

    def print_report(self, verbose: bool = False) -> None:
        """Prints the values of the Sterimol parameters.

        Args:
            verbose: Whether to print uncorrected L_value and bond length
        """
        if verbose:
            print(
                f"{'L':10s}{'B_1':10s}{'B_5':10s}" f"{'L_uncorr':10s}{'d(a1-a2)':10s}"
            )
            print(
                f"{self.L_value:<10.2f}{self.B_1_value:<10.2f}"
                f"{self.B_5_value:<10.2f}{self.L_value_uncorrected:<10.2f}"
                f"{self.bond_length:<10.2f}"
            )
        else:
            print(f"{'L':10s}{'B_1':10s}{'B_5':10s}")
            print(
                f"{self.L_value:<10.2f}{self.B_1_value:<10.2f}"
                f"{self.B_5_value:<10.2f}"
            )

    @requires_dependency(
        [
            Import(module="matplotlib.colors", item="hex2color"),
            Import(module="pyvista", alias="pv"),
            Import(module="pyvistaqt", item="BackgroundPlotter"),
        ],
        globals(),
    )
    def draw_3D(
        self,
        atom_scale: float = 0.5,
        background_color: str = "white",
        arrow_color: str = "steelblue",
    ) -> None:
        """Draw a 3D representation of the molecule with the Sterimol vectors.

        Args:
            atom_scale: Scaling factor for atom size
            background_color: Background color for plot
            arrow_color: Arrow color
        """
        # Set up plotter
        p = BackgroundPlotter()
        p.set_background(background_color)

        # Draw molecule
        for atom in self._atoms:
            color = hex2color(jmol_colors[atom.element])
            if atom.element == 0:
                radius = 0.5 * atom_scale
            else:
                radius = atom.radius * atom_scale
            sphere = pv.Sphere(center=list(atom.coordinates), radius=radius)
            if atom.index in self._excluded_atoms:
                opacity = 0.25
            else:
                opacity = 1
            p.add_mesh(sphere, color=color, opacity=opacity, name=str(atom.index))

        # Draw sphere for Buried Sterimol
        if hasattr(self, "_sphere_radius"):
            sphere = pv.Sphere(
                center=self._dummy_atom.coordinates, radius=self._sphere_radius
            )
            p.add_mesh(sphere, opacity=0.25)

        if hasattr(self, "_points"):
            p.add_points(self._points, color="gray")

        # Get arrow starting points
        start_L = self._dummy_atom.coordinates
        start_B = self._attached_atom.coordinates

        # Add L arrow with label
        length = np.linalg.norm(self.L)
        direction = self.L / length
        stop_L = start_L + length * direction
        L_arrow = get_drawing_arrow(start=start_L, direction=direction, length=length)
        p.add_mesh(L_arrow, color=arrow_color)

        # Add B_1 arrow
        length = np.linalg.norm(self.B_1)
        direction = self.B_1 / length
        stop_B_1 = start_B + length * direction
        B_1_arrow = get_drawing_arrow(start=start_B, direction=direction, length=length)
        p.add_mesh(B_1_arrow, color=arrow_color)

        # Add B_5 arrow
        length = np.linalg.norm(self.B_5)
        direction = self.B_5 / length
        stop_B_5 = start_B + length * direction
        B_5_arrow = get_drawing_arrow(start=start_B, direction=direction, length=length)
        p.add_mesh(B_5_arrow, color=arrow_color)

        # Add labels
        points = np.vstack([stop_L, stop_B_1, stop_B_5])
        labels = ["L", "B1", "B5"]
        p.add_point_labels(
            points,
            labels,
            text_color="black",
            font_size=30,
            bold=False,
            show_points=False,
            point_size=1,
        )

        self._plotter = p

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self._atoms)!r} atoms)"


def cli(file: str) -> Any:
    """CLI for Sterimol.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    elements, coordinates = read_geometry(file)
    return functools.partial(Sterimol, elements, coordinates)



"""Pyramidalization code."""

import functools
import itertools
from typing import Any, Iterable, List, Optional, Sequence, Union

import numpy as np
import scipy.spatial

from morfeus.io import read_geometry
from morfeus.utils import get_connectivity_matrix


class Pyramidalization:
    """Calculates and stores results of pyramidalization and alpha angle.

    As described in Struct. Chem. 1991, 2, 107 and alternatively according to bond
    angles as in J. Comput. Chem. 2012, 33 (27), 2173–2179.

    Args:
        coordinates: Coordinates (Å)
        atom_index: Index of pyramidalized atom (1-indexed)
        neighbor_indices: Indices of neighbors to pyramidalized atom
        elements: Elements as atomic symbols or numbers
        radii: Covalent radii used to determine connectivity (Å)
        radii_type: Covalent radii type: 'pyykko'
        excluded_atoms: Indices of atoms to exclude
        method: Method for detecting neighbors: 'connectivity' or 'distance'. Ignored if
            neighbor_indices is given.

    Attributes:
        alpha: Average alpha angle (degrees)
        alphas: Alpha angles for all permutations of neighbors (degrees)
        neighbor_indices: Indices of neighbors to pyramidalized atom
        P_angle: Pyramidalization according to Gavrish
        P: Pyramidalization according to Radhakrishnan
    """

    alpha: float
    alphas: np.ndarray
    neighbor_indices: List[int]
    P_angle: float
    P: float

    def __init__(  # noqa: C901
        self,
        coordinates: Sequence[Sequence[float]],
        atom_index: int,
        neighbor_indices: Optional[Sequence[int]] = None,
        elements: Optional[Union[Iterable[int], Iterable[str]]] = None,
        radii: Optional[Sequence[float]] = None,
        radii_type: str = "pyykko",
        excluded_atoms: Optional[Sequence[int]] = None,
        method: str = "distance",
        scale_factor: float = 1.2,
    ) -> None:
        coordinates = np.array(coordinates)
        atom_coordinates = coordinates[atom_index - 1]

        if neighbor_indices is None:
            neighbor_indices = []
        else:
            neighbor_indices = list(neighbor_indices)

        if excluded_atoms is None:
            excluded_atoms = []
        excluded_atoms = excluded_atoms = np.array(excluded_atoms, dtype=bool)

        # Get 3 closest neighbors
        if len(neighbor_indices) > 0:
            if len(neighbor_indices) != 3:
                raise Exception(f"Only {len(neighbor_indices)} neighbors.")
            neighbors = np.array(neighbor_indices) - 1
        elif method == "distance":
            # Generate mask for excluded atoms
            mask = np.zeros(len(coordinates), dtype=bool)
            mask[excluded_atoms - 1] = True
            mask[atom_index - 1] = True

            # Get three closest atoms not in the excluded atoms
            distances = scipy.spatial.distance.cdist(
                atom_coordinates.reshape(1, -1), coordinates
            ).reshape(-1)
            distances[mask] = np.inf
            neighbors = np.argsort(distances)[:3]
        elif method == "connectivity":
            # Construct connectivity matrix and get closest neighbors.
            if elements is None and radii is None:
                raise Exception("Connectivity requires elements or radii.")
            # if radii is None:
            #    radii = get_radii(elements, radii_type="pyykko")
            connectivity_matrix = get_connectivity_matrix(
                coordinates,
                elements=elements,
                radii=radii,
                radii_type=radii_type,
                scale_factor=scale_factor,
            )
            connected_atoms = np.where(connectivity_matrix[atom_index - 1, :])[0]
            neighbors = connected_atoms[~np.isin(connected_atoms, excluded_atoms - 1)]
            if len(neighbors) != 3:
                raise Exception(f"{len(neighbors)} neighbors. 3 expected.")

        # Get unit vectors between central atom and neighbors
        a = coordinates[neighbors[0]] - atom_coordinates
        a /= np.linalg.norm(a)
        b = coordinates[neighbors[1]] - atom_coordinates
        b /= np.linalg.norm(b)
        c = coordinates[neighbors[2]] - atom_coordinates
        c /= np.linalg.norm(c)

        # Calculate alpha for all permutations
        alphas = []
        vectors = []
        cos_alphas = []
        thetas = []
        for v_1, v_2, v_3 in itertools.permutations([a, b, c], 3):
            # Calculate cos_alpha
            normal = np.cross(v_1, v_2)
            normal /= np.linalg.norm(normal)
            cos_alpha = np.dot(v_3, normal)

            # Test if normal vector is colinear with v_3
            if cos_alpha < 0:
                continue
            alpha = np.arccos(cos_alpha)

            # Check for "acute" pyramid and correct angle
            v_1_2 = v_1 + v_2
            v_1_2 /= np.linalg.norm(v_1_2)
            cos_angle = np.dot(v_1_2, v_3)
            if cos_angle > 0:
                alpha = -alpha
            alphas.append(alpha)
            cos_alphas.append(cos_alpha)
            vectors.append((v_1, v_2))

            # Calculate theta angle
            cos_theta = np.dot(v_1, v_2)
            theta = np.rad2deg(np.arccos(cos_theta))
            thetas.append(theta)

        # Calculate P
        v_1, v_2 = vectors[0]
        sin_theta = np.linalg.norm(np.cross(v_1, v_2))
        P = sin_theta * cos_alphas[0]

        # Correct P if pyramid is "acute" on average
        if np.mean(alphas) < 0:
            P = 2 - P

        # Calculate P according to Gavrish method
        P_angle = np.sqrt(360 - sum(thetas))

        # Store attributes
        self.P = P
        self.P_angle = P_angle
        self.alpha = np.rad2deg(np.mean(alphas))
        self.alphas = np.rad2deg(alphas)
        self.neighbor_indices = (neighbors + 1).tolist()

    def print_report(self) -> None:
        """Print report of results."""
        print(f"P: {self.P:.3f}")
        print(f"P_angle: {self.P_angle:.3f}")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({round(self.P, 3)!r})"


def cli(file: str) -> Any:
    """CLI for pyramidalization.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    elements, coordinates = read_geometry(file)
    return functools.partial(Pyramidalization, coordinates, elements=elements)



"""Input and output."""

from os import PathLike
from pathlib import Path
import typing
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple, Union

import numpy as np

from morfeus.d3_data import r2_r4
from morfeus.data import BOHR_TO_ANGSTROM, KCAL_TO_HARTREE
from morfeus.typing import Array1D, Array2D, Array3D, ArrayLike2D, ArrayLike3D
from morfeus.utils import convert_elements, Import, requires_dependency

if typing.TYPE_CHECKING:
    import cclib.io


class CrestParser:
    """Parses folder with CREST output files for conformer information.

    Reads the files: cre_members, crest.energies, crest_conformers.xyz

    Args:
        path: Path to CREST folder

    Attributes:
        conformer_coordinates: Conformer coordinates (Å)
        degeneracies: Degeneracies
        elements: Elements
        energies: Energies (a.u.)
    """

    conformer_coordinates: Array3D
    degeneracies: Array1D
    elements: Array1D
    energies: Array1D

    def __init__(self, path: Union[str, PathLike]) -> None:
        path = Path(path)

        with open(path / "cre_members") as f:
            lines = f.readlines()
        degeneracies = []
        for line in lines:
            strip_line = line.strip().split()
            if len(strip_line) != 1:
                degeneracies.append(int(strip_line[0]))
        degeneracies = np.array(degeneracies)

        with open(path / "crest.energies") as f:
            lines = f.readlines()

        energies = []
        for line in lines:
            strip_line = line.strip().split()
            energies.append(float(strip_line[1]))
        energies = np.array(energies) * KCAL_TO_HARTREE

        elements, conformer_coordinates = read_xyz(path / "crest_conformers.xyz")

        self.elements = elements
        self.conformer_coordinates = conformer_coordinates
        self.energies = energies
        self.degeneracies = degeneracies


class CubeParser:
    """Parses Gaussian cube file of electron density.

    Args:
        file: Cube file

    Attributes:
        min_x: Minimum x value (Å)
        min_y: Minimum y value (Å)
        min_z: Minimum z value (Å)
        step_x: Step size in x direction (Å)
        step_y: Step size in y direction (Å)
        step_z: Step size in z direction (Å)
        X: 3D array of x values (Å)
        Y: 3D array of y values (Å)
        Z: 3D array of z values (Å)
        S: 3D array of electron density scalars (electorns / Bohr^3)
        n_points_x: Number of points in the x direction
        n_points_y: Number of points in the y direction
        n_points_z: Number of points in the < direction
    """

    min_x: float
    min_y: float
    min_z: float
    step_x: float
    step_y: float
    step_z: float
    X: Array3D
    Y: Array3D
    Z: Array3D
    S: Array3D
    n_points_x: int
    n_points_y: int
    n_points_z: int

    def __init__(self, file: Union[str, PathLike]) -> None:
        # Read the lines from the cube file
        with open(file) as file:
            lines = file.readlines()

        # Skip first two lines which are comments
        lines = lines[2:]

        # Get the number of atoms
        n_atoms = int(lines[0].strip().split()[0])

        # Get the minimum values along the axes
        min_x = float(lines[0].strip().split()[1]) * BOHR_TO_ANGSTROM
        min_y = float(lines[0].strip().split()[2]) * BOHR_TO_ANGSTROM
        min_z = float(lines[0].strip().split()[3]) * BOHR_TO_ANGSTROM

        # Get the number of points and step size along each axis
        n_points_x = int(lines[1].strip().split()[0])
        step_x = float(lines[1].strip().split()[1]) * BOHR_TO_ANGSTROM

        n_points_y = int(lines[2].strip().split()[0])
        step_y = float(lines[2].strip().split()[2]) * BOHR_TO_ANGSTROM

        n_points_z = int(lines[3].strip().split()[0])
        step_z = float(lines[3].strip().split()[3]) * BOHR_TO_ANGSTROM

        # Generate grid
        x = min_x + np.arange(0, n_points_x) * step_x
        y = min_y + np.arange(0, n_points_y) * step_y
        z = min_z + np.arange(0, n_points_z) * step_z
        X, Y, Z = np.meshgrid(x, y, z, indexing="ij")

        # Skips to data lines and read in data
        lines = lines[(n_atoms + 4) :]

        data = []
        for line in lines:
            line_data = [float(datum) for datum in line.strip().split()]
            data.extend(line_data)

        # Create array
        S = np.array(data).reshape(X.shape)

        # Set up attributes
        self.X = X
        self.Y = Y
        self.Z = Z
        self.S = S

        self.min_x = min_x
        self.min_y = min_y
        self.min_z = min_z

        self.step_x = step_x
        self.step_y = step_y
        self.step_z = step_z

        self.n_points_x = n_points_x
        self.n_points_y = n_points_y
        self.n_points_z = n_points_z

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.S.size!r} points)"


class D3Parser:
    """Parses the output of Grimme's D3 program.

    Extracts the C₆ᴬᴬ and C₈ᴬᴬ coefficients

    Args:
        file: File containing output from the D3 program.

    Attributes:
        c6_coefficients: C₆ᴬᴬ coefficients (a.u.)
        c8_coefficients: C₈ᴬᴬ coefficients (a.u.)
    """

    c6_coefficients: Array1D
    c8_coefficients: Array1D

    def __init__(self, file: Union[str, PathLike]) -> None:
        # Read the file
        with open(file, encoding="utf-8") as file:
            lines = file.readlines()

        # Parse the file for the coefficients
        c6_coefficients = []
        c8_coefficients = []
        read = False
        for line in lines:
            if read:
                if not line.strip():
                    read = False
                    break
                strip_line = line.strip().split()
                c6 = float(strip_line[7])
                c8 = float(strip_line[8])
                c6_coefficients.append(c6)
                c8_coefficients.append(c8)
            if "C8(AA)" in line:
                read = True

        # Set attributes
        self.c6_coefficients = np.array(c6_coefficients)
        self.c8_coefficients = np.array(c8_coefficients)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"


class D4Parser:
    """Parses the output of Grimme's D4 program.

    Extracts the C₆ᴬᴬ and C₈ᴬᴬ coefficients

    Args:
        file: File containing output from the D4 program.

    Attributes:
        c6_coefficients: C₆ᴬᴬ coefficients (a.u.)
        c8_coefficients: C₈ᴬᴬ coefficients (a.u.)
    """

    c6_coefficients: Array1D
    c8_coefficients: Array1D

    def __init__(self, file: Union[str, PathLike]) -> None:
        # Read the file
        with open(file, encoding="utf-8") as file:
            lines = file.readlines()

        # Parse the file and extract the coefficients
        c6_coefficients = []
        elements = []
        read = False
        for line in lines:
            if read:
                if not line.strip():
                    read = False
                    break
                strip_line = line.strip().split()
                element = int(strip_line[1])
                elements.append(element)
                c6 = float(strip_line[5])
                c6_coefficients.append(c6)
            if "C6AA" in line:
                read = True
        c6_coefficients = np.array(c6_coefficients)
        c8_coefficients = np.array(
            [
                3 * c6 * r2_r4[element] ** 2
                for c6, element in zip(c6_coefficients, elements)
            ]
        )

        # Set up attributes
        self.c6_coefficients = c6_coefficients
        self.c8_coefficients = c8_coefficients

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"


class VertexParser:
    """Parses the contents of a Multiwfn vtx.pdb file.

    Extracts the vertices and faces of the surface.

    Args:
        file: Name of file containing the vertices.

    Attributes:
        faces: Faces of surface
        vertices: Vertices of surface
    """

    faces: Optional[np.ndarray]
    vertices: np.ndarray

    def __init__(self, file: Union[str, PathLike]) -> None:  # noqa: C901
        # Parse file to see if it containts connectivity
        with open(file) as file:
            lines = file.readlines()

        # Get the number of vertices
        n_vertices = int(lines[0].strip().split()[5])

        # Parse the vertex positions and their connectivities
        vertices = {}
        vertex_map = {}
        connectivities: Dict[int, Set[int]] = {
            i: set() for i in range(1, n_vertices + 1)
        }
        vertex_counter = 1
        included_vertex_counter = 1
        for line in lines:
            if "HETATM" in line:
                if line[13] == "C":
                    x = float(line[32:39])
                    y = float(line[40:47])
                    z = float(line[48:55])
                    vertices[vertex_counter] = [x, y, z]
                    vertex_map[vertex_counter] = included_vertex_counter
                    included_vertex_counter += 1
                vertex_counter += 1
            if "CONECT" in line:
                n_entries = int(len(line.strip()) / 6 - 1)
                entries = []
                for i in range(1, n_entries + 1):
                    entry = int(line[i * 6 : i * 6 + 6])
                    entries.append(entry)
                connectivities[entries[0]].update(entries[1:])

        # Establish faces based on connectivity
        # https://stackoverflow.com/questions/1705824/finding-cycle-of-3-nodes-or-triangles-in-a-graph # noqa: B950
        if any(connectivities.values()):
            faces = []
            visited = set()
            for vertex_1 in connectivities:
                temp_visited = set()
                for vertex_2 in connectivities[vertex_1]:
                    if vertex_2 in visited:
                        continue
                    for vertex_3 in connectivities[vertex_2]:
                        if vertex_3 in visited or vertex_3 in temp_visited:
                            continue
                        if vertex_1 in connectivities[vertex_3]:
                            triangle_vertices = [vertex_1, vertex_2, vertex_3]
                            mapped_vertices = [
                                vertex_map[vertex] - 1 for vertex in triangle_vertices
                            ]
                            faces.append(mapped_vertices)
                    temp_visited.add(vertex_2)
                visited.add(vertex_1)
            self.faces = np.array(faces)
        else:
            self.faces = None

        # Set up attributes.
        self.vertices = np.array(list(vertices.values()))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self.vertices)!r} points)"


def read_gjf(file: Union[str, PathLike]) -> Tuple[Array1D, Array1D]:
    """Reads Gaussian gjf/com file and returns elements and coordinates.

    Args:
        file: gjf/com file object.

    Returns:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
    """
    # Read file and split lines
    with open(file) as f:
        text = f.read()
    xyz_lines = text.split("\n\n")[2].splitlines()[1:]
    xyz_lines = [line for line in xyz_lines if line.strip()]

    # Loop over lines and store elements and coordinates
    elements: List[Union[int, str]] = []
    coordinates: List[List[float]] = []
    for line in xyz_lines:
        split_line = line.strip().split()
        element = split_line[0]
        if element.isdigit():
            elements.append(int(element))
        else:
            elements.append(element)
        atom_coordinates = [float(i) for i in split_line[1:]]
        coordinates.append(atom_coordinates)
    elements = np.array(elements)
    coordinates = np.array(coordinates)

    return elements, coordinates


def read_geometry(
    file: Union[str, PathLike]
) -> Tuple[Array1D, Union[Array2D, Array3D]]:
    """Read geometry file and guess parser based on suffix.

    Args:
        file: Filename or Path object

    Returns:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)

    Raises:
        ValueError: When suffix is not supported
    """
    if isinstance(file, str):
        suffix = file.lower().split(".")[-1]
    elif isinstance(file, PathLike):
        path = Path(file)
        suffix = path.suffix[1:]

    if suffix == "xyz":
        elements, coordinates = read_xyz(file)
    elif suffix in ("gjf", "com"):
        elements, coordinates = read_gjf(file)
    else:
        raise ValueError(f"File suffix {suffix} not supported.")

    return elements, coordinates


def read_xyz(file: Union[str, PathLike]) -> Tuple[Array1D, Union[Array2D, Array3D]]:
    """Reads xyz file.

    Returns elements as written (atomic numbers or symbols) and coordinates.

    Args:
        file: Filename or Path object

    Returns:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
    """
    # Read file and split lines
    with open(file) as f:
        lines = f.readlines()

    # Loop over lines and store elements and coordinates
    elements: List[Union[int, str]] = []
    coordinates: List[List[float]] = []
    n_atoms = int(lines[0].strip())
    line_chunks = zip(*[iter(lines)] * (n_atoms + 2))
    for line_chunk in line_chunks:
        for line in line_chunk[2:]:
            strip_line = line.strip().split()
            atom = strip_line[0]
            if atom.isdigit():
                atom = int(atom)
            elements.append(atom)
            coordinates.append(
                [float(strip_line[1]), float(strip_line[2]), float(strip_line[3])]
            )
    elements = np.array(elements)[:n_atoms]
    coordinates = np.array(coordinates).reshape(-1, n_atoms, 3)
    if coordinates.shape[0] == 1:
        coordinates = coordinates[0]

    return elements, coordinates


@requires_dependency([Import("cclib")], globals())
def read_cclib(file: Any) -> Tuple[Array1D, Array1D]:
    """Reads geometry file with cclib.

    Returns elements as atomic numbers and coordinates.

    Args:
        file: Input file to cclib

    Returns:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
    """
    data = cclib.io.ccread(file)
    elements = data.atomnos
    coordinates = data.atomcoords

    if coordinates.shape[0] == 1:
        coordinates = coordinates[0]

    return elements, coordinates


def get_xyz_string(
    symbols: Sequence[str], coordinates: Sequence[Sequence[float]], comment: str = ""
) -> str:
    """Return xyz string.

    Args:
        symbols: Atomic symbols
        coordinates: Atomic coordinates (Å)
        comment: Comment

    Returns:
        string: XYZ string
    """
    string = f"{len(symbols)}\n"
    string += f"{comment}\n"
    for s, c in zip(symbols, coordinates):
        string += f"{s:10s}{c[0]:10.5f}{c[1]:10.5f}{c[2]:10.5f}\n"

    return string


def write_xyz(
    file: Union[str, PathLike],
    elements: Union[Iterable[int], Iterable[str]],
    coordinates: Union[ArrayLike2D, ArrayLike3D],
    comments: Optional[Sequence[str]] = None,
) -> None:
    """Writes xyz file from elements and coordinates.

    Args:
        file: xyz file or path object
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        comments: Comments
    """
    # Convert elements to symbols
    symbols = convert_elements(elements, output="symbols")
    coordinates = np.array(coordinates).reshape(-1, len(symbols), 3)
    if comments is None:
        comments = [""] * len(coordinates)

    # Write the xyz file
    with open(file, "w") as f:
        for coord, comment in zip(coordinates, comments):
            xyz_string = get_xyz_string(symbols, coord, comment=comment)
            f.write(xyz_string)
            f.write("\n")



"""xtb interface code."""

import functools
import typing
from typing import Any, Dict, Iterable, Optional, Union

import numpy as np

from morfeus.data import ANGSTROM_TO_BOHR, HARTREE_TO_EV
from morfeus.io import read_geometry
from morfeus.typing import Array1D, ArrayLike2D
from morfeus.utils import convert_elements, Import, requires_dependency

if typing.TYPE_CHECKING:
    import xtb
    import xtb.interface

IPEA_CORRECTIONS = {"1": 5.700, "2": 4.846}


@requires_dependency([Import("xtb"), Import("xtb.interface")], globals())
class XTB:
    """Calculates electronic properties with the xtb-python package.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        version: Version of xtb to use. Currently works with '1' or '2'.
        charge: Molecular charge
        n_unpaired: Number of unpaired electrons
        solvent: Solvent. See xtb-python documentation
        electronic_temperature: Electronic temperature (K)
    """

    _charge: int
    _coordinates: np.ndarray
    _electronic_temperature: Optional[int]
    _elements: np.ndarray
    _n_unpaired: Optional[int]
    _results: Any
    _solvent: Optional[str]
    _version: str

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        version: str = "2",
        charge: int = 0,
        n_unpaired: Optional[int] = None,
        solvent: Optional[str] = None,
        electronic_temperature: Optional[int] = None,
    ) -> None:
        # Converting elements to atomic numbers if the are symbols
        self._elements = np.array(convert_elements(elements, output="numbers"))

        # Store settings
        self._coordinates = np.array(coordinates)
        self._version = version
        self._charge = charge
        self._solvent = solvent
        self._n_unpaired = n_unpaired
        self._electronic_temperature = electronic_temperature

        # Set up results dictionary
        self._results = {
            -1: None,
            0: None,
            1: None,
        }

        self._params = {
            "1": xtb.interface.Param.GFN1xTB,
            "2": xtb.interface.Param.GFN2xTB,
        }

    def get_bond_order(self, i: int, j: int) -> float:
        """Returns bond order between two atoms.

        Args:
            i: Index of atom 1 (1-indexed)
            j: Index of atom 2 (1-indexed)

        Returns:
            bond_order: Bond order
        """
        bo_matrix = self.get_bond_orders()
        bond_order: float = bo_matrix[i - 1, j - 1]

        return bond_order

    def get_bond_orders(self, charge_state: int = 0) -> Array1D:
        """Returns bond orders.

        Args:
            charge_state: Charge state relative to original charge

        Returns:
            bond_orders: Bond orders
        """
        self._check_results(charge_state)
        bond_orders: np.ndarray = self._results[charge_state].get_bond_orders()

        return bond_orders

    def _get_charges(self, charge_state: int = 0) -> Array1D:
        """Returns atomic charges."""
        self._check_results(charge_state)
        charges: np.ndarray = self._results[charge_state].get_charges()

        return charges

    def get_charges(self, charge_state: int = 0) -> Dict[int, float]:
        """Returns atomic charges.

        Args:
            charge_state: Charge state relative to original charge

        Returns:
            charges: Atomic charges
        """
        self._check_results(charge_state)
        charges = self._results[charge_state].get_charges()
        charges = {i: charge for i, charge in enumerate(charges, start=1)}

        return charges

    def get_dipole(self, charge_state: int = 0) -> Array1D:
        """Calculate dipole vector (a.u.).

        Args:
            charge_state: Charge state relative to original charge

        Returns:
            dipole: Dipole vector
        """
        self._check_results(charge_state)
        dipole: np.ndarray = self._results[charge_state].get_dipole()

        return dipole

    def get_ea(self, corrected: bool = False) -> float:
        """Calculate electron affinity.

        Args:
            corrected: Whether to apply correction term

        Returns:
            ea: Electron affinity (eV)
        """
        # Calculate energies
        energy_neutral = self._get_energy(0)
        energy_anion = self._get_energy(-1)

        # Calculate electron affinity
        ea = (energy_neutral - energy_anion) * HARTREE_TO_EV
        if corrected:
            ea -= IPEA_CORRECTIONS[self._version]

        return ea

    def get_fukui(self, variety: str) -> Dict[int, float]:
        """Calculate Fukui coefficients.

        Args:
            variety: Type of Fukui coefficient: 'nucleophilicity', 'electrophilicity',
                'radical', 'dual', 'local_nucleophilicity' or 'local_electrophilicity'.

        Returns:
            fukui: Atomic Fukui coefficients

        Raises:
            ValueError: When variety does not exist
        """
        varieties = [
            "dual",
            "electrophilicity",
            "local_electrophilicity",
            "local_nucleophilicity",
            "nucleophilicity",
            "radical",
        ]
        fukui: np.ndarray
        if variety in ["local_nucleophilicity", "nucleophilicity"]:
            fukui = self._get_charges(0) - self._get_charges(1)
        elif variety == "electrophilicity":
            fukui = self._get_charges(-1) - self._get_charges(0)
        elif variety == "radical":
            fukui = (self._get_charges(-1) - self._get_charges(0)) / 2
        elif variety == "dual":
            fukui = (
                2 * self._get_charges(0) - self._get_charges(1) - self._get_charges(-1)
            )
        elif variety == "local_electrophilicity":
            fukui_radical = np.array(list(self.get_fukui("radical").values()))
            fukui_dual = np.array(list(self.get_fukui("dual").values()))
            chem_pot = -(self.get_ip() + self.get_ea()) / 2
            hardness = self.get_ip() - self.get_ea()
            fukui = (
                -(chem_pot / hardness) * fukui_radical
                + 1 / 2 * (chem_pot / hardness) ** 2 * fukui_dual
            )
        else:
            raise ValueError(
                f"Variety '{variety}' does not exist. "
                f"Choose one of {', '.join(varieties)}."
            )

        fukui = {i: fukui for i, fukui in enumerate(fukui, start=1)}

        return fukui

    def get_global_descriptor(self, variety: str, corrected: bool = False) -> float:
        """Calculate global reactivity descriptors.

        Args:
            corrected: Whether to apply correction term
            variety: Type of descriptor: 'electrophilicity', 'nucleophilicity',
                'electrofugality' or 'nucleofugality'

        Returns:
            descriptor: Global reactivity descriptor (eV)

        Raises:
            ValueError: When variety does not exist
        """
        varieties = [
            "electrofugality",
            "electrophilicity",
            "nucleofugality",
            "nucleophilicity",
        ]
        if variety == "electrophilicity":
            descriptor = (
                self.get_ip(corrected=corrected) + self.get_ea(corrected=corrected)
            ) ** 2 / (
                8
                * (self.get_ip(corrected=corrected) - self.get_ea(corrected=corrected))
            )
        elif variety == "nucleophilicity":
            descriptor = -self.get_ip(corrected=corrected)
        elif variety == "electrofugality":
            descriptor = (
                3 * self.get_ip(corrected=corrected) - self.get_ea(corrected=corrected)
            ) ** 2 / (
                8
                * (self.get_ip(corrected=corrected) - self.get_ea(corrected=corrected))
            )
        elif variety == "nucleofugality":
            descriptor = (
                self.get_ip(corrected=corrected) - 3 * self.get_ea(corrected=corrected)
            ) ** 2 / (
                8
                * (self.get_ip(corrected=corrected) - self.get_ea(corrected=corrected))
            )
        else:
            raise ValueError(
                f"Variety '{variety}' does not exist. "
                f"Choose one of {', '.join(varieties)}."
            )

        return descriptor

    def get_homo(self) -> float:
        """Calculate HOMO energy.

        Returns:
            homo_energy: HOMO energy (a.u.)
        """
        eigenvalues = self._get_eigenvalues()
        occupations = self._get_occupations()
        homo_index = int(occupations.sum().round(0) / 2) - 1
        homo_energy: float = eigenvalues[homo_index]

        return homo_energy

    def get_ip(self, corrected: bool = False) -> float:
        """Calculate ionization potential.

        Args:
            corrected: Whether to apply correction term

        Returns:
            ip: Ionization potential (eV)
        """
        # Calculate energies
        energy_neutral = self._get_energy(0)
        energy_cation = self._get_energy(1)

        # Calculate ionization potential
        ip = (energy_cation - energy_neutral) * HARTREE_TO_EV
        if corrected:
            ip -= IPEA_CORRECTIONS[self._version]

        return ip

    def get_lumo(self) -> float:
        """Calculate LUMO energy.

        Returns:
            lumo_energy: LUMO energy (a.u.)
        """
        eigenvalues = self._get_eigenvalues()
        occupations = self._get_occupations()
        homo_index = int(occupations.sum().round(0) / 2) - 1
        lumo_index = homo_index + 1
        lumo_energy: float = eigenvalues[lumo_index]

        return lumo_energy

    def _check_results(self, charge_state: int) -> None:
        """Checks whether results are already calculated and does calculation."""
        if self._results[charge_state] is None:
            self._sp(charge_state)

    def _get_eigenvalues(self) -> Array1D:
        """Get orbital eigenvalues."""
        self._check_results(0)
        eigenvalues: np.ndarray = self._results[0].get_orbital_eigenvalues()
        return eigenvalues

    def _get_energy(self, charge_state: int = 0) -> float:
        """Get total energy."""
        self._check_results(charge_state)
        energy: float = self._results[charge_state].get_energy()
        return energy

    def _get_occupations(self) -> Array1D:
        """Get occupation numbers."""
        self._check_results(0)
        occupations: np.ndarray = self._results[0].get_orbital_occupations()
        return occupations

    def _sp(self, charge_state: int = 0) -> None:
        """Perform single point calculation."""
        # Set up calculator
        calc = xtb.interface.Calculator(
            self._params[self._version],
            self._elements,
            self._coordinates * ANGSTROM_TO_BOHR,
            charge=self._charge + charge_state,
            uhf=self._n_unpaired,
        )
        calc.set_verbosity(xtb.libxtb.VERBOSITY_MUTED)

        # Set solvent
        if self._solvent:
            solvent = xtb.utils.get_solvent(self._solvent)
            if solvent is None:
                raise Exception(f"Solvent '{self._solvent}' not recognized")
            calc.set_solvent(solvent)

        # Set electronic temperature
        if self._electronic_temperature:
            calc.set_electronic_temperature(self._electronic_temperature)

        # Do singlepoint calculation and store the result
        res = calc.singlepoint()
        self._results[charge_state] = res


def cli(file: str) -> Any:
    """CLI for XTB.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    elements, coordinates = read_geometry(file)
    return functools.partial(XTB, elements, coordinates)



"""Typing code for arrays.

Borrowed from gpaw/typing.py
"""

from typing import Any

import numpy as np

try:
    from numpy.typing import ArrayLike
except ImportError:
    ArrayLike = Any

ArrayLike1D = ArrayLike
ArrayLike2D = ArrayLike
ArrayLike3D = ArrayLike

ArrayND = np.ndarray
Array1D = ArrayND
Array2D = ArrayND
Array3D = ArrayND



"""Calculates steric descriptors for molecules.

Modules:
    buried_volume: Buried volume code
    calculators: Internal calculators
    cone_angle: Cone angle code
    conformer: Conformer tools
    d3_data: Reference data for D3 calculations
    data: Module related to data
    geometry: Help classes and functions related to geometry
    io: Input and output
    local_force: Local force constant code
    plotting: Plotting functions
    pyramidalization: Pyramidalization code
    qc: Interface to quantum-chemical programs
    sasa: Solvent accessible surface area code
    sterimol: Sterimol code
    typing: Typing code for arrays
    utils: Helper functions.
    visible_volume: Visible volume code
    xtb: xtb interface code

Scripts:
    morfeus: Command line interface
"""
from importlib import metadata

from morfeus.buried_volume import BuriedVolume
from morfeus.cone_angle import ConeAngle
from morfeus.dispersion import Dispersion
from morfeus.io import read_geometry, read_gjf, read_xyz
from morfeus.local_force import LocalForce
from morfeus.pyramidalization import Pyramidalization
from morfeus.sasa import SASA
from morfeus.sterimol import Sterimol
from morfeus.visible_volume import VisibleVolume
from morfeus.xtb import XTB

__all__ = [
    "read_geometry",
    "read_gjf",
    "read_xyz",
    "BuriedVolume",
    "ConeAngle",
    "Dispersion",
    "LocalForce",
    "Pyramidalization",
    "SASA",
    "Sterimol",
    "VisibleVolume",
    "XTB",
]

# Version of the morfeus package
__version__ = metadata.version("morfeus-ml")



"""Help classes and functions related to geometry."""

import math
from typing import Iterable, List, Optional, Sequence, Union

import numpy as np
from scipy.spatial.transform import Rotation

from morfeus.data import ANGSTROM_TO_BOHR
from morfeus.typing import Array1D, Array2D, ArrayLike1D, ArrayLike2D
from morfeus.utils import get_connectivity_matrix


class Atom:
    """Atom common for morfeus calculations.

    Args:
        element: Atomic number (starting from 1)
        coordinates: Coordinates (Å)
        radius: vdW radius (Å)
        index: Atom index (starting from 1)

    Attributes:
        accessible_mask: Boolean mask for accessible points
        accessible_points: Points accessible to solvent (Å)
        area: Solvent-accessible surface area (Å²)
        cone: Cone tangent to atom
        coordinates: Coordinates (Å)
        coordination_number: Coordination number
        element: Atomic number (1-indexed)
        index: Atom index (1-indexed)
        invisible_mask: Boolean mask for invisible points
        occluded_mask: Boolean mask for occluded points
        occluded_points: Points occluded by other atoms (Å)
        p_values: P values
        point_areas: Point areas (Å²)
        point_volumes: Point volumes (Å³)
        point: Points (Å)
        proximal_mask: Boolean mask for proximal points
        radius: vdW radius (Å)
        volume: Volume inside solvent-accessible surface area (Å³)
    """

    accessible_mask: Array1D
    accessible_points: Array2D
    area: float
    cone: "Cone"
    coordinates: Array1D
    coordination_number: float
    element: int
    index: int
    invisible_mask: Array1D
    occluded_mask: Array1D
    occluded_points: Array2D
    p_values: Array1D
    point_areas: Array1D
    point_volumes: Array1D
    points: Array2D
    proximal_mask: Array1D
    radius: float
    volume: float

    def __init__(
        self, element: int, coordinates: ArrayLike1D, radius: float, index: int
    ) -> None:
        # Set up initial attributes
        self.coordinates = np.array(coordinates)
        self.element = element
        self.index = index
        self.radius = radius

    def get_cone(self) -> None:
        """Construct cone for atom."""
        vector = self.coordinates
        normal = vector / np.linalg.norm(vector)
        sin_alpha = self.radius / np.linalg.norm(vector)
        alpha = math.asin(sin_alpha)
        cone = Cone(alpha, [self], normal)
        self.cone = cone

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.index!r})"


class Cone:
    """Cone used in cone angle calculations.

    Args:
        angle: Cone angle (rad)
        atoms: Atoms that are tangent to cone (1-indexed)
        normal: Normal vector of cone

    Attributes:
        angle: Cone angle (rad)
        atoms: Atoms that are tangent to cone (1-indexed)
        normal: Normal vector of cone
    """

    angle: float
    atoms: List[Atom]
    normal: Array1D

    def __init__(
        self, angle: float, atoms: Sequence[Atom], normal: ArrayLike1D
    ) -> None:
        self.angle = angle
        self.atoms = list(atoms)
        self.normal = np.array(normal)

    def is_inside(self, atom: Atom) -> bool:
        """Tests if atom lies inside the cone.

        Args:
            atom: Atom to test.

        Returns:
            True if inside, False if outside.
        """
        # Get vertex angle of atom
        beta = atom.cone.angle

        # Calculate angle between cone normal vector and unit vector to atom
        cos = np.dot(atom.cone.normal, self.normal)

        # Take into account numerical problems that sometimes gives a value
        # somewhat above 1
        if 1 - cos > 0 and 1 - cos < 1e-5:
            cos = 1
        angle = math.acos(cos)

        # Check if atom lies inside cone, within numerical reason
        diff = self.angle - (beta + angle)
        if diff > -1e-5:
            return True
        else:
            return False

    def is_inside_points(self, points: ArrayLike2D, method: str = "cross") -> Array1D:
        """Test if points are inside cone of atom.

        Args:
            points: Points to check (Å)
            method: Method for testing: 'angle', 'cross' or 'dot'

        Returns:
            is_inside: Boolean array with points marked as inside

        Raises:
            ValueError: When method not supported
        """
        points = np.array(points)
        if method in ["cross", "dot"]:
            # Calculate radius of cone at distance of each point
            cone_distances = np.dot(points, self.normal)
            cone_radii = np.tan(self.angle) * cone_distances

            # Calculate orthogonal distance of points to cone normal vector
            if method == "cross":
                orth_distances = np.linalg.norm(np.cross(-points, self.normal), axis=1)
            elif method == "dot":
                orth_distances = np.linalg.norm(
                    -points - np.dot(points, self.normal).reshape(-1, 1) * self.normal,
                    axis=1,
                )

            # Determine if distance is smaller than cone radius.
            inside = orth_distances < cone_radii
        elif method == "angle":
            norm_points = points / np.linalg.norm(points, axis=1).reshape(-1, 1)

            # Calculate angle between cone normal vector and unit vector to
            # atom
            cos = np.dot(norm_points, self.normal)

            # Take into account numerical problems that sometimes gives a value
            # somewhat above 1
            cos[np.logical_and(1 - cos > 0, 1 - cos < 1e-5)] = 1
            angle = np.arccos(cos)

            # Check if atom lies inside cone, within numerical reason
            diff = self.angle - angle

            inside = diff > -1e-5
        else:
            raise ValueError(f"method={method} not supported.")

        is_inside: np.ndarray = inside

        return is_inside

    def __repr__(self) -> str:
        atoms = ", ".join([str(atom.index) for atom in self.atoms])
        return f"{self.__class__.__name__}(Tangent atoms: {atoms})"


class Sphere:
    """Sphere class for creating and holding points on vdW surface.

    Args:
        center: Coordinates for center (Å)
        density: Area per point (Å²) for empty sphere
            and volume per point (Å³) for filled sphere.
        filled: Whether a sphere with internal points should be constructed (works only
            with method='projection')
        method: Method for generating points: 'fibonacci', 'polar' or 'projection'
        radius: Radius (Å)

    Attributes:
        area: Area (Å²)
        center: Coordinates for sphere center (Å)
        circumference: Circumference (Å)
        density: Density of points (Å² or Å³)
        points: Points in/on sphere (Å)
        radius: Radius (Å)
        volume: Volume (Å³)
    """

    area: float
    center: Array1D
    circumference: float
    density: float
    points: Array1D
    radius: float
    volume: float

    def __init__(
        self,
        center: ArrayLike1D,
        radius: float,
        density: float = 0.005,
        method: str = "fibonacci",
        filled: bool = False,
    ) -> None:
        self.center = np.array(center)
        self.radius = radius
        self.circumference = math.pi * radius * 2
        self.area = 4 * radius ** 2 * math.pi
        self.volume = 4 * radius ** 3 * math.pi / 3
        self.density = density

        if method == "polar":
            self.points = self._get_points_polar(density=density)
        elif method == "projection":
            self.points = self._get_points_projected(density=density, filled=filled)
        elif method == "fibonacci":
            self.points = self._get_points_fibonacci(density=density)

    @staticmethod
    def _get_cartesian_coordinates(
        r: float, theta: ArrayLike1D, phi: ArrayLike1D
    ) -> Array2D:
        """Converts polar to Cartesian coordinates.

        Args:
            phi: Phi angles (radians)
            r: Radius (Å)
            theta: Theta angles (radians)

        Returns:
            points: Cartesian points (Å)
        """
        # Calculate x, y and z coordinates
        x = r * np.sin(theta) * np.cos(phi)
        y = r * np.sin(theta) * np.sin(phi)
        z = r * np.cos(theta)

        # Stack coordinates as columns
        points: np.ndarray = np.column_stack((x, y, z))

        return points

    def _get_points_fibonacci(self, density: float) -> Array2D:
        """Construct points on sphere surface by the Fibonacci golden spiral method.

        Method vectorized from
        https://stackoverflow.com/questions/9600801/evenly-distributing-n-points-on-a-sphere

        Args:
            density: Area per point (Å²)

        Returns:
            points: Surface points (Å)
        """
        # Generate points on unit sphere
        rnd = 1
        n = int(round((self.area / density)))
        offset = 2.0 / n
        increment = math.pi * (3.0 - math.sqrt(5.0))

        i = np.arange(n)
        y = ((i * offset) - 1) + (offset / 2)
        r = np.sqrt(1 - np.square(y))
        phi = np.mod((i + rnd), n) * increment
        x = np.cos(phi) * r
        z = np.sin(phi) * r

        # Generate points and adjust with radius and center
        points = np.column_stack((x, y, z))
        points = points * self.radius
        points: np.ndarray = points + self.center

        return points

    def _get_points_polar(self, density: float) -> Array2D:
        """Construct points on sphere by polar coordinate method.

        Args:
            density: Area per point (Å²)

        Returns:
            points: Points on sphere (Å)
        """
        # Calculate number of points
        n = int(round((self.area / density / 2) ** (1 / 2)))

        # Set up points along theta and phi
        theta = np.linspace(0, math.pi, n)
        phi = np.linspace(0, 2 * math.pi, 2 * n)

        # Combine together all the possible combinations of theta and phi
        combined_theta_phi = np.dstack(np.meshgrid(theta, phi)).reshape(-1, 2)
        theta = combined_theta_phi[:, 0]
        phi = combined_theta_phi[:, 1]

        # Get the Cartesian coordinates
        points = self._get_cartesian_coordinates(self.radius, theta, phi)

        # Adjust to sphere center
        points: np.ndarray = points + self.center

        return points

    def _get_points_projected(self, density: float, filled: bool = False) -> Array2D:
        """Construct points on sphere surface by projection.

        Args:
            density: Area per point (Å²) for empty sphere and volume
                per point (Å³) for filled sphere
            filled: Whether to generate internal points

        Returns:
            points: Array of surface points (Å)
        """
        # Calculate number of points from density of empty or filled sphere.
        if filled:
            n = int(round((self.volume / density * 6 / math.pi) ** (1 / 3)))
        else:
            n = int(round((self.area / density * 6 / math.pi) ** (1 / 3)))

        # Generate points in box
        r = self.radius
        x = np.linspace(-r, r, n)
        y = np.linspace(-r, r, n)
        z = np.linspace(-r, r, n)
        points = np.stack(np.meshgrid(x, y, z), -1).reshape(-1, 3)

        # Remove points outside of sphere
        lengths = np.linalg.norm(points, axis=1)
        points = points[lengths <= r]

        # Project points onto sphere surface if sphere is empty
        if not filled:
            points = points / np.linalg.norm(points, axis=1).reshape(-1, 1) * r

        # Adjust with sphere center
        points: np.ndarray = points + self.center

        return points

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(center: {self.center}, "
            f"radius: {self.radius})"
        )


class Bond:
    """Bond internal coordinate.

    Args:
        atom_1: Index of atom 1 (1-indexed)
        atom_2: Index of atom 2 (1-indexed)

    Attributes:
        i: Index of atom 1 (1-indexed)
        j: Index of atom 2 (1-indexed)
    """

    i: int
    j: int
    atoms: List[int]

    def __init__(self, atom_1: int, atom_2: int) -> None:
        self.i = atom_1
        self.j = atom_2
        self.atoms = [atom_1, atom_2]

    def get_b_vector(self, coordinates: ArrayLike2D) -> Array1D:
        """Calculate vector of B matrix for internal coordinate.

        Code adapted from pyberny: https://github.com/jhrmnn/pyberny.

        Args:
            coordinates: Coordinates (Å)

        Returns:
            b_vector: Vector of B matrix (a.u.)
        """
        coordinates = np.array(coordinates)
        i, j = self.i - 1, self.j - 1
        v = (coordinates[i] - coordinates[j]) * ANGSTROM_TO_BOHR
        r = np.linalg.norm(v)
        grad = np.array([v / r, -v / r])

        b_vector = np.zeros(coordinates.size)
        b_vector[i * 3 : i * 3 + 3] = grad[0]
        b_vector[j * 3 : j * 3 + 3] = grad[1]

        return b_vector

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return (self.i == other.i or self.i == other.j) and (
                self.j == other.i or self.j == other.j
            )
        else:
            return False

    def __hash__(self) -> int:
        return hash((self.i, self.j))

    def __repr__(self) -> str:
        atoms = ", ".join(str(i) for i in sorted(self.atoms))
        return f"{self.__class__.__name__}({atoms})"


class Angle:
    """Bond internal coordinate.

    Code adapted from pyberny: https://github.com/jhrmnn/pyberny.

    Args:
        atom_1: Index of atom 1 (1-indexed)
        atom_2: Index of atom 2 (1-indexed)
        atom_3: Index of atom 3 (1-indexed)

    Attributes:
        i: Index of atom 1 (1-indexed)
        j: Index of atom 2 (1-indexed)
        k: Index of atom 3 (1-indexed)
    """

    i: int
    j: int
    k: int
    atoms: List[int]

    def __init__(self, atom_1: int, atom_2: int, atom_3: int) -> None:
        self.i = atom_1
        self.j = atom_2
        self.k = atom_3
        self.atoms = [atom_1, atom_2, atom_3]

    def get_b_vector(self, coordinates: ArrayLike2D) -> Array1D:
        """Calculate vector of B matrix for internal coordinate.

        Code adapted from pyberny: https://github.com/jhrmnn/pyberny.

        Args:
            coordinates: Coordinates (Å)

        Returns:
            b_vector: Vector of B matrix (a.u.)
        """
        coordinates = np.array(coordinates)
        i, j, k = self.i - 1, self.j - 1, self.k - 1
        v_1 = (coordinates[i] - coordinates[j]) * ANGSTROM_TO_BOHR
        v_2 = (coordinates[k] - coordinates[j]) * ANGSTROM_TO_BOHR
        dot_product = np.dot(v_1, v_2) / (np.linalg.norm(v_1) * np.linalg.norm(v_2))
        if dot_product < -1:
            dot_product = -1
        elif dot_product > 1:
            dot_product = 1
        phi = np.arccos(dot_product)
        if abs(phi) > np.pi - 1e-6:
            grad = [
                (np.pi - phi) / (2 * np.linalg.norm(v_1) ** 2) * v_1,
                (1 / np.linalg.norm(v_1) - 1 / np.linalg.norm(v_2))
                * (np.pi - phi)
                / (2 * np.linalg.norm(v_1))
                * v_1,
                (np.pi - phi) / (2 * np.linalg.norm(v_2) ** 2) * v_2,
            ]
        else:
            grad = [
                1 / np.tan(phi) * v_1 / np.linalg.norm(v_1) ** 2
                - v_2 / (np.linalg.norm(v_1) * np.linalg.norm(v_2) * np.sin(phi)),
                (v_1 + v_2) / (np.linalg.norm(v_1) * np.linalg.norm(v_2) * np.sin(phi))
                - 1
                / np.tan(phi)
                * (v_1 / np.linalg.norm(v_1) ** 2 + v_2 / np.linalg.norm(v_2) ** 2),
                1 / np.tan(phi) * v_2 / np.linalg.norm(v_2) ** 2
                - v_1 / (np.linalg.norm(v_1) * np.linalg.norm(v_2) * np.sin(phi)),
            ]

        b_vector = np.zeros(coordinates.size)
        b_vector[i * 3 : i * 3 + 3] = grad[0]
        b_vector[j * 3 : j * 3 + 3] = grad[1]
        b_vector[k * 3 : k * 3 + 3] = grad[2]

        return b_vector

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return (
                (self.i == other.i or self.i == other.k)
                and (self.k == other.i or self.k == other.k)
                and self.j == other.j
            )
        else:
            return False

    def __hash__(self) -> int:
        return hash((self.i, self.j, self.k))

    def __repr__(self) -> str:
        atoms = ", ".join(str(i) for i in sorted(self.atoms))
        return f"{self.__class__.__name__}({atoms})"


class Dihedral:
    """Bond internal coordinate.

    Code adapted from pyberny: https://github.com/jhrmnn/pyberny.

    Args:
        atom_1: Index of atom 1 (1-indexed)
        atom_2: Index of atom 2 (1-indexed)
        atom_3: Index of atom 3 (1-indexed)
        atom_4: Index of atom 4 (1-indexed)

    Attributes:
        i: Index of atom 1 (1-indexed)
        j: Index of atom 2 (1-indexed)
        k: Index of atom 3 (1-indexed)
        l: Index of atom 4 (1-indexed)
    """

    i: int
    j: int
    k: int
    l: int
    atoms: List[int]

    def __init__(self, atom_1: int, atom_2: int, atom_3: int, atom_4: int) -> None:
        self.i = atom_1
        self.j = atom_2
        self.k = atom_3
        self.l = atom_4
        self.atoms = [atom_1, atom_2, atom_3, atom_4]

    def get_b_vector(self, coordinates: ArrayLike2D) -> Array1D:
        """Calculate vector of B matrix for internal coordinate.

        Code adapted from pyberny: https://github.com/jhrmnn/pyberny.

        Args:
            coordinates: Coordinates (Å)

        Returns:
            b_vector: Vector of B matrix (a.u.)
        """
        coordinates = np.array(coordinates)
        i, j, k, l = self.i - 1, self.j - 1, self.k - 1, self.l - 1
        v_1 = (coordinates[i] - coordinates[j]) * ANGSTROM_TO_BOHR
        v_2 = (coordinates[l] - coordinates[k]) * ANGSTROM_TO_BOHR
        w = (coordinates[k] - coordinates[j]) * ANGSTROM_TO_BOHR
        ew = w / np.linalg.norm(w)
        a_1 = v_1 - np.dot(v_1, ew) * ew
        a_2 = v_2 - np.dot(v_2, ew) * ew
        sgn = np.sign(np.linalg.det(np.array([v_2, v_1, w])))
        sgn = sgn or 1
        dot_product = np.dot(a_1, a_2) / (np.linalg.norm(a_1) * np.linalg.norm(a_2))
        if dot_product < -1:
            dot_product = -1
        elif dot_product > 1:
            dot_product = 1
        phi = np.arccos(dot_product) * sgn

        if abs(phi) > np.pi - 1e-6:
            g = np.cross(w, a_1)
            g = g / np.linalg.norm(g)
            A = np.dot(v_1, ew) / np.linalg.norm(w)
            B = np.dot(v_2, ew) / np.linalg.norm(w)
            grad = [
                g / (np.linalg.norm(g) * np.linalg.norm(a_1)),
                -((1 - A) / np.linalg.norm(a_1) - B / np.linalg.norm(a_2)) * g,
                -((1 + B) / np.linalg.norm(a_2) + A / np.linalg.norm(a_1)) * g,
                g / (np.linalg.norm(g) * np.linalg.norm(a_2)),
            ]
        elif abs(phi) < 1e-6:
            g = np.cross(w, a_1)
            g = g / np.linalg.norm(g)
            A = np.dot(v_1, ew) / np.linalg.norm(w)
            B = np.dot(v_2, ew) / np.linalg.norm(w)
            grad = [
                g / (np.linalg.norm(g) * np.linalg.norm(a_1)),
                -((1 - A) / np.linalg.norm(a_1) + B / np.linalg.norm(a_2)) * g,
                ((1 + B) / np.linalg.norm(a_2) - A / np.linalg.norm(a_1)) * g,
                -g / (np.linalg.norm(g) * np.linalg.norm(a_2)),
            ]
        else:
            A = np.dot(v_1, ew) / np.linalg.norm(w)
            B = np.dot(v_2, ew) / np.linalg.norm(w)
            grad = [
                1 / np.tan(phi) * a_1 / np.linalg.norm(a_1) ** 2
                - a_2 / (np.linalg.norm(a_1) * np.linalg.norm(a_2) * np.sin(phi)),
                ((1 - A) * a_2 - B * a_1)
                / (np.linalg.norm(a_1) * np.linalg.norm(a_2) * np.sin(phi))
                - 1
                / np.tan(phi)
                * (
                    (1 - A) * a_1 / np.linalg.norm(a_1) ** 2
                    - B * a_2 / np.linalg.norm(a_2) ** 2
                ),
                ((1 + B) * a_1 + A * a_2)
                / (np.linalg.norm(a_1) * np.linalg.norm(a_2) * np.sin(phi))
                - 1
                / np.tan(phi)
                * (
                    (1 + B) * a_2 / np.linalg.norm(a_2) ** 2
                    + A * a_1 / np.linalg.norm(a_1) ** 2
                ),
                1 / np.tan(phi) * a_2 / np.linalg.norm(a_2) ** 2
                - a_1 / (np.linalg.norm(a_1) * np.linalg.norm(a_2) * np.sin(phi)),
            ]
        b_vector = np.zeros(coordinates.size)
        b_vector[i * 3 : i * 3 + 3] = grad[0]
        b_vector[j * 3 : j * 3 + 3] = grad[1]
        b_vector[k * 3 : k * 3 + 3] = grad[2]
        b_vector[l * 3 : l * 3 + 3] = grad[3]

        return b_vector

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return all(
                [
                    self.i == other.i,
                    self.j == other.j,
                    self.k == other.k,
                    self.l == other.l,
                ]
            ) or all(
                [
                    self.i == other.l,
                    self.j == other.k,
                    self.k == other.j,
                    self.l == other.i,
                ]
            )
        else:
            return False

    def __hash__(self) -> int:
        return hash((self.i, self.j, self.k, self.l))

    def __repr__(self) -> str:
        atoms = ", ".join(str(i) for i in sorted(self.atoms))
        return f"{self.__class__.__name__}({atoms})"


class InternalCoordinates:
    """Internal coordinates.

    Attributes:
        internal_coordinates: Internal coordinates
    """

    internal_coordinates: List[Union[Bond, Angle, Dihedral]]

    def __init__(self) -> None:
        self.internal_coordinates = []

    def add_bond(self, atom_1: int, atom_2: int) -> None:
        """Add bonds to internal coordinates.

        Args:
            atom_1: Index of atom 1 (1-indexed)
            atom_2: Index of atom 2 (1-indexed)
        """
        bond = Bond(atom_1, atom_2)
        if bond not in self.internal_coordinates:
            self.internal_coordinates.append(bond)

    def add_angle(self, atom_1: int, atom_2: int, atom_3: int) -> None:
        """Add angle to internal coordinates.

        Args:
            atom_1: Index of atom 1 (1-indexed)
            atom_2: Index of atom 2 (1-indexed)
            atom_3: Index of atom 3 (1-indexed)
        """
        angle = Angle(atom_1, atom_2, atom_3)
        if angle not in self.internal_coordinates:
            self.internal_coordinates.append(angle)

    def add_dihedral(self, atom_1: int, atom_2: int, atom_3: int, atom_4: int) -> None:
        """Add dihedral angle to internal coordinates.

        Args:
            atom_1: Index of atom 1 (1-indexed)
            atom_2: Index of atom 2 (1-indexed)
            atom_3: Index of atom 3 (1-indexed)
            atom_4: Index of atom 4 (1-indexed)
        """
        dihedral = Dihedral(atom_1, atom_2, atom_3, atom_4)
        if dihedral not in self.internal_coordinates:
            self.internal_coordinates.append(dihedral)

    def add_internal_coordinate(self, atoms: Sequence[int]) -> None:
        """Add internal coordinate automatically depending on number of atoms.

        Args:
            atoms: Sequence of atom indices (1-index).
        """
        if len(atoms) == 2:
            self.add_bond(*atoms)
        elif len(atoms) == 3:
            self.add_angle(*atoms)
        elif len(atoms) == 4:
            self.add_dihedral(*atoms)

    def detect_bonds(
        self,
        coordinates: ArrayLike2D,
        elements: Optional[Union[Iterable[int], Iterable[str]]],
        radii: Optional[ArrayLike1D] = None,
        radii_type: str = "pyykko",
        scale_factor: float = 1.2,
    ) -> None:
        """Detect bonds based on covalent radii cutoffs."""
        # Detect bonds based on covalent radii
        connectivity_matrix = get_connectivity_matrix(
            coordinates,
            elements=elements,
            radii=radii,
            radii_type=radii_type,
            scale_factor=scale_factor,
        )
        indices = np.where(connectivity_matrix)

        bonds = set()
        for i, j in zip(*indices):
            bond = frozenset([i + 1, j + 1])
            bonds.add(bond)

        # Add each bond as an internal coordinate
        for bond in bonds:
            i, j = sorted(bond)
            self.add_bond(i, j)

    def get_B_matrix(self, coordinates: ArrayLike2D) -> Array2D:
        """Calculate B matrix for coordinates.

        Args:
            coordinates: Coordinates (Å)

        Returns:
            B_matrix: B matrix
        """
        coordinates = np.array(coordinates)
        b_vectors = []
        for internal_coordinate in self.internal_coordinates:
            b_vectors.append(internal_coordinate.get_b_vector(coordinates))
        B_matrix = np.vstack(b_vectors)

        return B_matrix

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({len(self.internal_coordinates)} "
            "coordinates)"
        )


def rotate_coordinates(
    coordinates: ArrayLike2D,
    vector: ArrayLike1D,
    axis: ArrayLike1D,
) -> Array2D:
    """Rotates coordinates by the rotation that aligns vector with axis.

    Args:
        axis: Reference vector to align input vector with
        coordinates: Coordinates to rotate
        vector: Vector to align

    Returns:
        rotated_coordinates: Rotated coordinates.
    """
    coordinates = np.array(coordinates)
    vector = np.array(vector)
    axis = np.array(axis)

    # Get real part of quaternion
    real = np.dot(vector, axis).reshape(1) + 1

    # Get imaginary dimensions and handle case of antiparallel vectors
    if real < 1e-6:
        w = np.cross(vector, np.array([0, 0, 1]))
        if np.linalg.norm(w) < 1e-6:
            w = np.cross(vector, np.array([1, 0, 0]))
    else:
        w = np.cross(vector, axis)

    # Form quaternion and normalize
    q = np.concatenate((w, real))
    q = q / np.linalg.norm(q)

    # Rotate atomic coordinates
    rotation = Rotation.from_quat(q)
    rotated_coordinates: np.ndarray = rotation.apply(coordinates)

    return rotated_coordinates


def kabsch_rotation_matrix(
    P: ArrayLike2D, Q: ArrayLike2D, center: bool = True
) -> Array2D:
    """Construct Kabsch rotation matrix.

    Constructs the rotation matrix that overlays the points in P with the points in Q
    with minimum RMSD. https://en.wikipedia.org/wiki/Kabsch_algorithm

    Args:
        P: Coordinates to be rotated
        Q: Reference coordinates
        center: Whether to center P and Q at origin.

    Returns:
        R: Rotation matrix
    """
    P = np.array(P)
    Q = np.array(Q)

    # Calculate centroids and center coordinates
    if center:
        centroid_P = np.mean(P, axis=0)
        centroid_Q = np.mean(Q, axis=0)
        P -= centroid_P
        Q -= centroid_Q

    # Calculate cross-covariance matrix
    H = P.T @ Q

    # Perform SVD
    U, S, V_T = np.linalg.svd(H)

    # Determine correction for right-hand coordinate system
    d = np.sign(np.linalg.det(V_T.T @ U.T))

    # Obtain rotation matrix
    R: np.ndarray = V_T.T @ np.array([[1, 0, 0], [0, 1, 0], [0, 0, d]]) @ U.T

    return R


def sphere_line_intersection(
    vector: ArrayLike1D, center: ArrayLike1D, radius: float
) -> Array2D:
    """Get points of intersection between line and sphere.

    Follows the procedure outlined here: http://paulbourke.net/geometry/circlesphere/.

    Args:
        vector: Vector giving direction of line
        center: Center of sphere
        radius: Radius of sphere

    Returns:
        intersection_points: Intersection points
    """
    vector = np.array(vector)
    center = np.array(center)

    # Set up points
    p_1 = vector
    p_2 = vector * 2

    # Set up terms for quadratic equation
    a = (p_2[0] - p_1[0]) ** 2 + (p_2[1] - p_1[1]) ** 2 + (p_2[2] - p_1[2]) ** 2
    b = 2 * (
        (p_2[0] - p_1[0]) * (p_1[0] - center[0])
        + (p_2[1] - p_1[1]) * (p_1[1] - center[1])
        + (p_2[2] - p_1[2]) * (p_1[2] - center[2])
    )
    c = (
        center[0] ** 2
        + center[1] ** 2
        + center[2] ** 2
        + p_1[0] ** 2
        + p_1[1] ** 2
        + p_1[2] ** 2
        - 2 * (center[0] * p_1[0] + center[1] * p_1[1] + center[2] * p_1[2])
        - radius ** 2
    )

    # Determine value within the square root and select cases
    within_sqrt = b ** 2 - 4 * a * c
    if within_sqrt < 0:
        us = []
    elif within_sqrt == 0:
        us = [-b / (2 * a)]
    elif within_sqrt > 0:
        us = [
            (-b + math.sqrt(within_sqrt)) / (2 * a),
            (-b - math.sqrt(within_sqrt)) / (2 * a),
        ]

    # Calculate intersection points.
    intersection_points = np.vstack([p_1 + u * (p_2 - p_1) for u in us])

    return intersection_points
